#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from gel import _testbase as tb
import edgedb

import array
import math
import unittest


# An array.array subtype where indexing doesn't work.
# We use this to verify that the non-boxing memoryview based
# fast path works, since the slow path won't work on this object.
class brokenarray(array.array):
    def __getitem__(self, i):
        raise AssertionError("the fast path wasn't used!")


class TestVector(tb.SyncQueryTestCase):

    PGVECTOR_VER = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.PGVECTOR_VER = cls.client.query_single('''
            select assert_single((
              select sys::ExtensionPackage filter .name = 'pgvector'
            )).version
        ''')

        if cls.PGVECTOR_VER is None:
            raise unittest.SkipTest("feature not implemented")

        cls.client.execute('''
            create extension pgvector;
        ''')

    @classmethod
    def tearDownClass(cls):
        try:
            cls.client.execute('''
                drop extension pgvector;
            ''')
        finally:
            super().tearDownClass()

    def test_vector_01(self):
        val = self.client.query_single('''
            select <ext::pgvector::vector>[1.5,2.0,3.8]
        ''')
        self.assertTrue(isinstance(val, array.array))
        self.assertEqual(val, array.array('f', [1.5, 2.0, 3.8]))

        val = self.client.query_single(
            '''
                select <json><ext::pgvector::vector>$0
            ''',
            [3.0, 9.0, -42.5],
        )
        self.assertEqual(val, '[3, 9, -42.5]')

        val = self.client.query_single(
            '''
                select <json><ext::pgvector::vector>$0
            ''',
            array.array('f', [3.0, 9.0, -42.5])
        )
        self.assertEqual(val, '[3, 9, -42.5]')

        val = self.client.query_single(
            '''
                select <json><ext::pgvector::vector>$0
            ''',
            array.array('i', [1, 2, 3]),
        )
        self.assertEqual(val, '[1, 2, 3]')

        # Test that the fast-path works: if the encoder tries to
        # call __getitem__ on this brokenarray, it will fail.
        val = self.client.query_single(
            '''
                select <json><ext::pgvector::vector>$0
            ''',
            brokenarray('f', [3.0, 9.0, -42.5])
        )
        self.assertEqual(val, '[3, 9, -42.5]')

        # I don't think it's worth adding a dependency to test this,
        # but this works too:
        # import numpy as np
        # val = self.client.query_single(
        #     '''
        #     select <json><ext::pgvector::vector>$0
        #     ''',
        #     np.asarray([3.0, 9.0, -42.5], dtype=np.float32),
        # )
        # self.assertEqual(val, '[3,9,-42.5]')

        # Some sad path tests
        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::vector>$0
                ''',
                [3.0, None, -42.5],
            )

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::vector>$0
                ''',
                [3.0, 'x', -42.5],
            )

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::vector>$0
                ''',
                'foo',
            )

    def test_vector_02(self):
        if self.PGVECTOR_VER < (0, 7):
            self.skipTest("need at least pgvector 0.7.4 for sparsevec")

        val = self.client.query_single(
            '''
            select <ext::pgvector::sparsevec>
                <ext::pgvector::vector>[0, 1.5, 2.0, 3.8, 0, 0]
            ''',
        )
        self.assertEqual(val['dim'], 6)
        self.assertEqual(val[1], 1.5)
        self.assertEqual(val[2], 2)
        self.assertTrue(math.isclose(val[3], 3.8, abs_tol=1e-6))

        val = self.client.query_single(
            '''
            select <array<float32>><ext::pgvector::vector>
                <ext::pgvector::sparsevec>$0
            ''',
            {'dim': 6, 1: 1.5, 2: 2, 3: 3.8},
        )
        self.assertEqual(len(val), 6)
        self.assertEqual(val[0], 0)
        self.assertEqual(val[1], 1.5)
        self.assertEqual(val[2], 2)
        self.assertTrue(math.isclose(val[3], 3.8, abs_tol=1e-6))
        self.assertEqual(val[4], 0)
        self.assertEqual(val[5], 0)

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::sparsevec>$0
                ''',
                {'dim': 1, 1: 1.5, 2: 2, 3: 3.8},
            )

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::sparsevec>$0
                ''',
                {'dims': 1, 1: 1.5, 2: 2, 3: 3.8},
            )

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::sparsevec>$0
                ''',
                {'dim': 6, 1: 1.5, 2: 2, 3: '3.8'},
            )

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::sparsevec>$0
                ''',
                {'dim': 6, 1: 1.5, 2: 2, '3': 3.8},
            )

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::sparsevec>$0
                ''',
                {'dim': 6, 1: 1.5, 2: 2, 3: 0},
            )

    def test_vector_03(self):
        if self.PGVECTOR_VER < (0, 7):
            self.skipTest("need at least pgvector 0.7.4 for halfvec")

        val = self.client.query_single(
            '''
            select <ext::pgvector::halfvec>
                [1.5, 2.0, 3.8, 0, 3.4575e-3, 65000,
                 6.0975e-5, 2.2345e-7, -5.96e-8]
            ''',
        )

        self.assertTrue(isinstance(val, array.array))
        self.assertEqual(val[0], 1.5)
        self.assertEqual(val[1], 2)
        self.assertTrue(math.isclose(val[2], 3.80, rel_tol=1e-3))
        self.assertEqual(val[3], 0)
        self.assertTrue(math.isclose(val[4], 3.457e-3, rel_tol=1e-3))
        self.assertTrue(math.isclose(val[5], 65000, rel_tol=1e-3))
        # These values are sub-normal so they don't map perfectly onto f32
        self.assertTrue(math.isclose(val[6], 6.0975e-5, rel_tol=1e-2))
        self.assertTrue(math.isclose(val[7], 2.38e-7, rel_tol=1e-2))
        self.assertTrue(math.isclose(val[8], -5.96e-8, rel_tol=1e-2))

        val = self.client.query_single(
            '''
            select <array<float32>><ext::pgvector::halfvec>$0
            ''',
            [1.5, 2.0, 3.8, 0, 3.4575e-3, 65000,
             6.0975e-5, 2.385e-7, -5.97e-8],
        )

        self.assertEqual(val[0], 1.5)
        self.assertEqual(val[1], 2)
        self.assertTrue(math.isclose(val[2], 3.80, rel_tol=1e-3))
        self.assertEqual(val[3], 0)
        self.assertTrue(math.isclose(val[4], 3.457e-3, rel_tol=1e-3))
        self.assertTrue(math.isclose(val[5], 65000, rel_tol=1e-3))
        # These values are sub-normal so they don't map perfectly onto f32
        self.assertTrue(math.isclose(val[6], 6.0975e-5, rel_tol=1e-2))
        self.assertTrue(math.isclose(val[7], 2.38e-7, rel_tol=1e-2))
        self.assertTrue(math.isclose(val[8], -5.96e-8, rel_tol=1e-2))

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::halfvec>$0
                ''',
                [3.0, None, -42.5],
            )

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::halfvec>$0
                ''',
                [3.0, 'x', -42.5],
            )

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::halfvec>$0
                ''',
                'foo',
            )

        with self.assertRaises(edgedb.InvalidArgumentError):
            self.client.query_single(
                '''
                    select <ext::pgvector::halfvec>$0
                ''',
                [1_000_000],
            )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import asyncio
import contextlib
import logging
import socket
import ssl
import typing

from . import abstract
from . import base_client
from . import con_utils
from . import errors
from . import transaction
from .protocol import asyncio_proto
from .protocol.protocol import InputLanguage, OutputFormat


__all__ = (
    'create_async_client', 'AsyncIOClient'
)


logger = logging.getLogger(__name__)


class AsyncIOConnection(base_client.BaseConnection):
    __slots__ = ("_loop",)

    def __init__(self, loop, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._loop = loop

    def is_closed(self):
        protocol = self._protocol
        return protocol is None or not protocol.connected

    async def connect_addr(self, addr, timeout):
        try:
            await asyncio.wait_for(self._connect_addr(addr), timeout)
        except asyncio.TimeoutError as e:
            raise TimeoutError from e

    async def sleep(self, seconds):
        await asyncio.sleep(seconds)

    async def aclose(self):
        """Send graceful termination message wait for connection to drop."""
        if not self.is_closed():
            try:
                self._protocol.terminate()
                await self._protocol.wait_for_disconnect()
            except (Exception, asyncio.CancelledError):
                self.terminate()
                raise
            finally:
                self._cleanup()

    def _protocol_factory(self):
        return asyncio_proto.AsyncIOProtocol(self._params, self._loop)

    async def _connect_addr(self, addr):
        tr = None

        try:
            if isinstance(addr, str):
                # UNIX socket
                tr, pr = await self._loop.create_unix_connection(
                    self._protocol_factory, addr
                )
            else:
                try:
                    tr, pr = await self._loop.create_connection(
                        self._protocol_factory,
                        *addr,
                        ssl=self._params.ssl_ctx,
                        server_hostname=(
                            self._params.tls_server_name or addr[0]
                        ),
                    )
                except ssl.CertificateError as e:
                    raise con_utils.wrap_error(e) from e
                except ssl.SSLError as e:
                    raise con_utils.wrap_error(e) from e
                else:
                    con_utils.check_alpn_protocol(
                        tr.get_extra_info('ssl_object')
                    )
        except socket.gaierror as e:
            # All name resolution errors are considered temporary
            raise errors.ClientConnectionFailedTemporarilyError(str(e)) from e
        except OSError as e:
            raise con_utils.wrap_error(e) from e
        except Exception:
            if tr is not None:
                tr.close()
            raise

        pr.set_connection(self)

        try:
            await pr.connect()
        except OSError as e:
            if tr is not None:
                tr.close()
            raise con_utils.wrap_error(e) from e
        except BaseException:
            if tr is not None:
                tr.close()
            raise

        self._protocol = pr
        self._addr = addr

    def _dispatch_log_message(self, msg):
        for cb in self._log_listeners:
            self._loop.call_soon(cb, self, msg)


class _PoolConnectionHolder(base_client.PoolConnectionHolder):
    __slots__ = ()
    _event_class = asyncio.Event

    async def close(self, *, wait=True):
        if self._con is None:
            return
        if wait:
            await self._con.aclose()
        else:
            self._pool._loop.create_task(self._con.aclose())

    async def wait_until_released(self, timeout=None):
        await self._release_event.wait()


class _AsyncIOPoolImpl(base_client.BasePoolImpl):
    __slots__ = ('_loop',)
    _holder_class = _PoolConnectionHolder

    def __init__(
        self,
        connect_args,
        *,
        max_concurrency: typing.Optional[int],
        connection_class,
    ):
        if not issubclass(connection_class, AsyncIOConnection):
            raise TypeError(
                f'connection_class is expected to be a subclass of '
                f'gel.asyncio_client.AsyncIOConnection, '
                f'got {connection_class}')
        self._loop = None
        super().__init__(
            connect_args,
            lambda *args: connection_class(self._loop, *args),
            max_concurrency=max_concurrency,
        )

    def _ensure_initialized(self):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()
            self._queue = asyncio.LifoQueue(maxsize=self._max_concurrency)
            self._first_connect_lock = asyncio.Lock()
            self._resize_holder_pool()

    def _set_queue_maxsize(self, maxsize):
        self._queue._maxsize = maxsize

    async def _maybe_get_first_connection(self):
        async with self._first_connect_lock:
            if self._working_addr is None:
                return await self._get_first_connection()

    async def acquire(self, timeout=None):
        self._ensure_initialized()

        async def _acquire_impl():
            ch = await self._queue.get()  # type: _PoolConnectionHolder
            try:
                proxy = await ch.acquire()  # type: AsyncIOConnection
            except (Exception, asyncio.CancelledError):
                self._queue.put_nowait(ch)
                raise
            else:
                # Record the timeout, as we will apply it by default
                # in release().
                ch._timeout = timeout
                return proxy

        if self._closing:
            raise errors.InterfaceError('pool is closing')

        if timeout is None:
            return await _acquire_impl()
        else:
            return await asyncio.wait_for(
                _acquire_impl(), timeout=timeout)

    async def _release(self, holder):

        if not isinstance(holder._con, AsyncIOConnection):
            raise errors.InterfaceError(
                f'release() received invalid connection: '
                f'{holder._con!r} does not belong to any connection pool'
            )

        timeout = None

        # Use asyncio.shield() to guarantee that task cancellation
        # does not prevent the connection from being returned to the
        # pool properly.
        return await asyncio.shield(holder.release(timeout))

    async def aclose(self):
        """Attempt to gracefully close all connections in the pool.

        Wait until all pool connections are released, close them and
        shut down the pool.  If any error (including cancellation) occurs
        in ``close()`` the pool will terminate by calling
        _AsyncIOPoolImpl.terminate() .

        It is advisable to use :func:`python:asyncio.wait_for` to set
        a timeout.
        """
        if self._closed:
            return

        if not self._loop:
            self._closed = True
            return

        self._closing = True

        try:
            warning_callback = self._loop.call_later(
                60, self._warn_on_long_close)

            release_coros = [
                ch.wait_until_released() for ch in self._holders]
            await asyncio.gather(*release_coros)

            close_coros = [
                ch.close() for ch in self._holders]
            await asyncio.gather(*close_coros)

        except (Exception, asyncio.CancelledError):
            self.terminate()
            raise

        finally:
            warning_callback.cancel()
            self._closed = True
            self._closing = False

    def _warn_on_long_close(self):
        logger.warning(
            'AsyncIOClient.aclose() is taking over 60 seconds to complete. '
            'Check if you have any unreleased connections left. '
            'Use asyncio.wait_for() to set a timeout for '
            'AsyncIOClient.aclose().')


class AsyncIOIteration(transaction.BaseTransaction, abstract.AsyncIOExecutor):

    __slots__ = ("_managed", "_locked")

    def __init__(self, retry, client, iteration):
        super().__init__(retry, client, iteration)
        self._managed = False
        self._locked = False

    async def __aenter__(self):
        if self._managed:
            raise errors.InterfaceError(
                'cannot enter context: already in an `async with` block')
        self._managed = True
        return self

    async def __aexit__(self, extype, ex, tb):
        with self._exclusive():
            self._managed = False
            return await self._exit(extype, ex)

    async def _ensure_transaction(self):
        if not self._managed:
            raise errors.InterfaceError(
                "Only managed retriable transactions are supported. "
                "Use `async with transaction:`"
            )
        await super()._ensure_transaction()

    async def _query(self, query_context: abstract.QueryContext):
        with self._exclusive():
            return await super()._query(query_context)

    async def _execute(self, execute_context: abstract.ExecuteContext) -> None:
        with self._exclusive():
            await super()._execute(execute_context)

    @contextlib.contextmanager
    def _exclusive(self):
        if self._locked:
            raise errors.InterfaceError(
                "concurrent queries within the same transaction "
                "are not allowed"
            )
        self._locked = True
        try:
            yield
        finally:
            self._locked = False


class AsyncIORetry(transaction.BaseRetry):

    def __aiter__(self):
        return self

    async def __anext__(self):
        # Note: when changing this code consider also
        # updating Retry.__next__.
        if self._done:
            raise StopAsyncIteration
        if self._next_backoff:
            await asyncio.sleep(self._next_backoff)
        self._done = True
        iteration = AsyncIOIteration(self, self._owner, self._iteration)
        self._iteration += 1
        return iteration


class AsyncIOClient(base_client.BaseClient, abstract.AsyncIOExecutor):
    """A lazy connection pool.

    A Client can be used to manage a set of connections to the database.
    Connections are first acquired from the pool, then used, and then released
    back to the pool.  Once a connection is released, it's reset to close all
    open cursors and other resources *except* prepared statements.

    Clients are created by calling
    :func:`~gel.asyncio_client.create_async_client`.
    """

    __slots__ = ()
    _impl_class = _AsyncIOPoolImpl

    async def ensure_connected(self):
        await self._impl.ensure_connected()
        return self

    async def aclose(self):
        """Attempt to gracefully close all connections in the pool.

        Wait until all pool connections are released, close them and
        shut down the pool.  If any error (including cancellation) occurs
        in ``aclose()`` the pool will terminate by calling
        AsyncIOClient.terminate() .

        It is advisable to use :func:`python:asyncio.wait_for` to set
        a timeout.
        """
        await self._impl.aclose()

    def transaction(self) -> AsyncIORetry:
        return AsyncIORetry(self)

    async def __aenter__(self):
        return await self.ensure_connected()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.aclose()

    async def _describe_query(
        self,
        query: str,
        *,
        inject_type_names: bool = False,
        input_language: InputLanguage = InputLanguage.EDGEQL,
        output_format: OutputFormat = OutputFormat.BINARY,
        expect_one: bool = False,
    ) -> abstract.DescribeResult:
        return await self._describe(abstract.DescribeContext(
            query=query,
            state=self._get_state(),
            inject_type_names=inject_type_names,
            input_language=input_language,
            output_format=output_format,
            expect_one=expect_one,
        ))


def create_async_client(
    dsn=None,
    *,
    max_concurrency=None,
    host: str = None,
    port: int = None,
    credentials: str = None,
    credentials_file: str = None,
    user: str = None,
    password: str = None,
    secret_key: str = None,
    database: str = None,
    branch: str = None,
    tls_ca: str = None,
    tls_ca_file: str = None,
    tls_security: str = None,
    wait_until_available: int = 30,
    timeout: int = 10,
):
    return AsyncIOClient(
        connection_class=AsyncIOConnection,
        max_concurrency=max_concurrency,

        # connect arguments
        dsn=dsn,
        host=host,
        port=port,
        credentials=credentials,
        credentials_file=credentials_file,
        user=user,
        password=password,
        secret_key=secret_key,
        database=database,
        branch=branch,
        tls_ca=tls_ca,
        tls_ca_file=tls_ca_file,
        tls_security=tls_security,
        wait_until_available=wait_until_available,
        timeout=timeout,
    )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2022-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from gel import _testbase as tb

from edgedb import State


class TestState(tb.TestCase):
    def test_state_default(self):
        self.assertEqual(State.defaults().as_dict(), {})

    def test_state_default_module(self):
        s1 = State.defaults()
        s2 = s1.with_default_module("m2")
        s3 = s2.with_default_module("m3")
        s4 = s3.with_default_module()

        self.assertEqual(s1.as_dict(), {})
        self.assertEqual(s2.as_dict(), {"module": "m2"})
        self.assertEqual(s3.as_dict(), {"module": "m3"})
        self.assertEqual(s4.as_dict(), {})

    def test_state_module_aliaases(self):
        s1 = State.defaults().with_default_module("m")
        s2 = s1.with_module_aliases(x="a", m2="i")
        s3 = s2.with_module_aliases({"x": "b", "m3": "j"})
        s4 = s3.without_module_aliases("x", "m3")
        s5 = s3.without_module_aliases()

        self.assertEqual(s1.as_dict(), {"module": "m"})
        self.assertEqual(s2.as_dict()["module"], "m")
        self.assertListEqual(
            s2.as_dict()["aliases"], [("x", "a"), ("m2", "i")]
        )
        self.assertEqual(s3.as_dict()["module"], "m")
        self.assertListEqual(
            s3.as_dict()["aliases"],
            [("x", "b"), ("m2", "i"), ("m3", "j")],
        )
        self.assertEqual(s4.as_dict()["module"], "m")
        self.assertListEqual(s4.as_dict()["aliases"], [("m2", "i")])
        self.assertEqual(s5.as_dict(), {"module": "m"})

    def test_state_config(self):
        s1 = State.defaults().with_default_module("m")
        s2 = s1.with_config(x="a", m2="i")
        s3 = s2.with_config({"x": "b", "m3": "j"})
        s4 = s3.without_config("x", "m3")
        s5 = s3.without_config()

        self.assertDictEqual(s1.as_dict(), {"module": "m"})
        self.assertEqual(s2.as_dict()["module"], "m")
        self.assertDictEqual(s2.as_dict()["config"], {"x": "a", "m2": "i"})
        self.assertEqual(s3.as_dict()["module"], "m")
        self.assertDictEqual(
            s3.as_dict()["config"],
            {"x": "b", "m2": "i", "m3": "j"},
        )
        self.assertEqual(s4.as_dict()["module"], "m")
        self.assertDictEqual(s4.as_dict()["config"], {"m2": "i"})
        self.assertEqual(s5.as_dict(), {"module": "m"})

    def test_state_globals(self):
        s1 = (
            State.defaults()
            .with_default_module("m")
            .with_module_aliases(a="x", b="y")
        )
        s2 = s1.with_globals({"a::g2": "22"}, i=2)
        s3 = s2.with_globals({"i": 3, "b::g3": "33"})
        s4 = s3.without_globals("i", "y::g3")
        s5 = s3.without_globals()

        self.assertDictEqual(
            s1.as_dict(), {"module": "m", "aliases": [("a", "x"), ("b", "y")]}
        )
        self.assertEqual(s2.as_dict()["module"], "m")
        self.assertListEqual(s2.as_dict()["aliases"], [("a", "x"), ("b", "y")])
        self.assertDictEqual(
            s2.as_dict()["globals"], {"m::i": 2, "x::g2": "22"}
        )
        self.assertEqual(s3.as_dict()["module"], "m")
        self.assertListEqual(s3.as_dict()["aliases"], [("a", "x"), ("b", "y")])
        self.assertDictEqual(
            s3.as_dict()["globals"], {"m::i": 3, "x::g2": "22", "y::g3": "33"}
        )
        self.assertEqual(s4.as_dict()["module"], "m")
        self.assertListEqual(s4.as_dict()["aliases"], [("a", "x"), ("b", "y")])
        self.assertDictEqual(s4.as_dict()["globals"], {"x::g2": "22"})
        self.assertDictEqual(
            s5.as_dict(), {"module": "m", "aliases": [("a", "x"), ("b", "y")]}
        )

        self.assertDictEqual(
            s3.with_default_module("x")
            .with_globals(g2="2222")
            .as_dict()["globals"],
            {"m::i": 3, "x::g2": "2222", "y::g3": "33"},
        )
        self.assertDictEqual(
            s3.with_default_module("x")
            .without_globals("g2")
            .without_module_aliases()
            .with_module_aliases(n="y")
            .with_default_module("y")
            .with_globals({"m::i": 4}, g3="3333")
            .as_dict()["globals"],
            {"m::i": 4, "y::g3": "3333"},
        )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2024-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations
import typing

import gel
import httpx
import httpx_sse

from . import types


def create_rag_client(client: gel.Client, **kwargs) -> RAGClient:
    client.ensure_connected()
    return RAGClient(client, types.RAGOptions(**kwargs))


async def create_async_rag_client(
    client: gel.AsyncIOClient, **kwargs
) -> AsyncRAGClient:
    await client.ensure_connected()
    return AsyncRAGClient(client, types.RAGOptions(**kwargs))


class BaseRAGClient:
    options: types.RAGOptions
    context: types.QueryContext
    client_cls = NotImplemented

    def __init__(
        self,
        client: typing.Union[gel.Client, gel.AsyncIOClient],
        options: types.RAGOptions,
        **kwargs,
    ):
        pool = client._impl
        host, port = pool._working_addr
        params = pool._working_params
        proto = "http" if params.tls_security == "insecure" else "https"
        branch = params.branch
        self.options = options
        self.context = types.QueryContext(**kwargs)
        args = dict(
            base_url=f"{proto}://{host}:{port}/branch/{branch}/ext/ai",
            verify=params.ssl_ctx,
        )
        if params.password is not None:
            args["auth"] = (params.user, params.password)
        elif params.secret_key is not None:
            args["headers"] = {"Authorization": f"Bearer {params.secret_key}"}
        self._init_client(**args)

    def _init_client(self, **kwargs):
        raise NotImplementedError

    def with_config(self, **kwargs) -> typing.Self:
        cls = type(self)
        rv = cls.__new__(cls)
        rv.options = self.options.derive(kwargs)
        rv.context = self.context
        rv.client = self.client
        return rv

    def with_context(self, **kwargs) -> typing.Self:
        cls = type(self)
        rv = cls.__new__(cls)
        rv.options = self.options
        rv.context = self.context.derive(kwargs)
        rv.client = self.client
        return rv

    def _make_rag_request(
        self,
        *,
        message: str,
        context: typing.Optional[types.QueryContext] = None,
        stream: bool,
    ) -> types.RAGRequest:
        if context is None:
            context = self.context
        return types.RAGRequest(
            model=self.options.model,
            prompt=self.options.prompt,
            context=context,
            query=message,
            stream=stream,
        )


class RAGClient(BaseRAGClient):
    client: httpx.Client

    def _init_client(self, **kwargs):
        self.client = httpx.Client(**kwargs)

    def query_rag(
        self, message: str, context: typing.Optional[types.QueryContext] = None
    ) -> str:
        resp = self.client.post(
            **self._make_rag_request(
                context=context,
                message=message,
                stream=False,
            ).to_httpx_request()
        )
        resp.raise_for_status()
        return resp.json()["response"]

    def stream_rag(
        self, message: str, context: typing.Optional[types.QueryContext] = None
    ) -> typing.Iterator[str]:
        with httpx_sse.connect_sse(
            self.client,
            "post",
            **self._make_rag_request(
                context=context,
                message=message,
                stream=True,
            ).to_httpx_request(),
        ) as event_source:
            event_source.response.raise_for_status()
            for sse in event_source.iter_sse():
                yield sse.data

    def generate_embeddings(self, *inputs: str, model: str) -> list[float]:
        resp = self.client.post(
            "/embeddings", json={"input": inputs, "model": model}
        )
        resp.raise_for_status()
        return resp.json()["data"][0]["embedding"]


class AsyncRAGClient(BaseRAGClient):
    client: httpx.AsyncClient

    def _init_client(self, **kwargs):
        self.client = httpx.AsyncClient(**kwargs)

    async def query_rag(
        self, message: str, context: typing.Optional[types.QueryContext] = None
    ) -> str:
        resp = await self.client.post(
            **self._make_rag_request(
                context=context,
                message=message,
                stream=False,
            ).to_httpx_request()
        )
        resp.raise_for_status()
        return resp.json()["response"]

    async def stream_rag(
        self, message: str, context: typing.Optional[types.QueryContext] = None
    ) -> typing.Iterator[str]:
        async with httpx_sse.aconnect_sse(
            self.client,
            "post",
            **self._make_rag_request(
                context=context,
                message=message,
                stream=True,
            ).to_httpx_request(),
        ) as event_source:
            event_source.response.raise_for_status()
            async for sse in event_source.aiter_sse():
                yield sse.data

    async def generate_embeddings(
        self, *inputs: str, model: str
    ) -> list[float]:
        resp = await self.client.post(
            "/embeddings", json={"input": inputs, "model": model}
        )
        resp.raise_for_status()
        return resp.json()["data"][0]["embedding"]



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import typing

import enum

from . import abstract
from . import errors
from . import options


class TransactionState(enum.Enum):
    NEW = 0
    STARTED = 1
    COMMITTED = 2
    ROLLEDBACK = 3
    FAILED = 4


class BaseTransaction:

    __slots__ = (
        '_client',
        '_connection',
        '_options',
        '_state',
        '__retry',
        '__iteration',
        '__started',
    )

    def __init__(self, retry, client, iteration):
        self._client = client
        self._connection = None
        self._options = retry._options.transaction_options
        self._state = TransactionState.NEW
        self.__retry = retry
        self.__iteration = iteration
        self.__started = False

    def is_active(self) -> bool:
        return self._state is TransactionState.STARTED

    def __check_state_base(self, opname):
        if self._state is TransactionState.COMMITTED:
            raise errors.InterfaceError(
                'cannot {}; the transaction is already committed'.format(
                    opname))
        if self._state is TransactionState.ROLLEDBACK:
            raise errors.InterfaceError(
                'cannot {}; the transaction is already rolled back'.format(
                    opname))
        if self._state is TransactionState.FAILED:
            raise errors.InterfaceError(
                'cannot {}; the transaction is in error state'.format(
                    opname))

    def __check_state(self, opname):
        if self._state is not TransactionState.STARTED:
            if self._state is TransactionState.NEW:
                raise errors.InterfaceError(
                    'cannot {}; the transaction is not yet started'.format(
                        opname))
            self.__check_state_base(opname)

    def _make_start_query(self):
        self.__check_state_base('start')
        if self._state is TransactionState.STARTED:
            raise errors.InterfaceError(
                'cannot start; the transaction is already started')

        return self._options.start_transaction_query()

    def _make_commit_query(self):
        self.__check_state('commit')
        return 'COMMIT;'

    def _make_rollback_query(self):
        self.__check_state('rollback')
        return 'ROLLBACK;'

    def __repr__(self):
        attrs = []
        attrs.append('state:{}'.format(self._state.name.lower()))
        attrs.append(repr(self._options))

        if self.__class__.__module__.startswith('gel.'):
            mod = 'gel'
        else:
            mod = self.__class__.__module__

        return '<{}.{} {} {:#x}>'.format(
            mod, self.__class__.__name__, ' '.join(attrs), id(self))

    async def _ensure_transaction(self):
        if not self.__started:
            self.__started = True
            query = self._make_start_query()
            self._connection = await self._client._impl.acquire()
            if self._connection.is_closed():
                await self._connection.connect(
                    single_attempt=self.__iteration != 0
                )
            try:
                await self._privileged_execute(query)
            except BaseException:
                self._state = TransactionState.FAILED
                raise
            else:
                self._state = TransactionState.STARTED

    async def _exit(self, extype, ex):
        if not self.__started:
            return False

        try:
            if extype is None:
                query = self._make_commit_query()
                state = TransactionState.COMMITTED
            else:
                query = self._make_rollback_query()
                state = TransactionState.ROLLEDBACK
            try:
                await self._privileged_execute(query)
            except BaseException:
                self._state = TransactionState.FAILED
                if extype is None:
                    # COMMIT itself may fail; recover in connection
                    await self._privileged_execute("ROLLBACK;")
                raise
            else:
                self._state = state
        except errors.EdgeDBError as err:
            if ex is None:
                # On commit we don't know if commit is succeeded before the
                # database have received it or after it have been done but
                # network is dropped before we were able to receive a response.
                # On a TransactionError, though, we know the we need
                # to retry.
                # TODO(tailhook) should other errors have retries?
                if (
                    isinstance(err, errors.TransactionError)
                    and err.has_tag(errors.SHOULD_RETRY)
                    and self.__retry._retry(err)
                ):
                    pass
                else:
                    raise err
            # If we were going to rollback, look at original error
            # to find out whether we want to retry, regardless of
            # the rollback error.
            # In this case we ignore rollback issue as original error is more
            # important, e.g. in case `CancelledError` it's important
            # to propagate it to cancel the whole task.
            # NOTE: rollback error is always swallowed, should we use
            # on_log_message for it?
        finally:
            await self._client._impl.release(self._connection)

        if (
            extype is not None and
            issubclass(extype, errors.EdgeDBError) and
            ex.has_tag(errors.SHOULD_RETRY)
        ):
            return self.__retry._retry(ex)

    def _get_query_cache(self) -> abstract.QueryCache:
        return self._client._get_query_cache()

    def _get_retry_options(self) -> typing.Optional[options.RetryOptions]:
        return None

    def _get_state(self) -> options.State:
        return self._client._get_state()

    def _get_warning_handler(self) -> options.WarningHandler:
        return self._client._get_warning_handler()

    def _get_annotations(self) -> typing.Dict[str, str]:
        return self._client._get_annotations()

    async def _query(self, query_context: abstract.QueryContext):
        await self._ensure_transaction()
        return await self._connection.raw_query(query_context)

    async def _execute(self, execute_context: abstract.ExecuteContext) -> None:
        await self._ensure_transaction()
        await self._connection._execute(execute_context)

    async def _privileged_execute(self, query: str) -> None:
        await self._connection.privileged_execute(abstract.ExecuteContext(
            query=abstract.QueryWithArgs(query, (), {}),
            cache=self._get_query_cache(),
            state=self._get_state(),
            retry_options=self._get_retry_options(),
            warning_handler=self._get_warning_handler(),
            annotations=self._get_annotations(),
        ))


class BaseRetry:

    def __init__(self, owner):
        self._owner = owner
        self._iteration = 0
        self._done = False
        self._next_backoff = 0
        self._options = owner._options

    def _retry(self, exc):
        self._last_exception = exc
        rule = self._options.retry_options.get_rule_for_exception(exc)
        if self._iteration >= rule.attempts:
            return False
        self._done = False
        self._next_backoff = rule.backoff(self._iteration)
        return True



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2019-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio
import queue
import random
import threading
import time

import edgedb

from gel import _testbase as tb
from edgedb import errors
from edgedb import blocking_client


class TestBlockingClient(tb.SyncQueryTestCase):
    def create_client(self, **kwargs):
        conargs = self.get_connect_args().copy()
        conargs["database"] = self.get_database_name()
        conargs["timeout"] = 120
        conargs.update(kwargs)
        conargs.setdefault(
            "connection_class", blocking_client.BlockingIOConnection
        )
        conargs.setdefault("max_concurrency", None)

        return tb.TestClient(**conargs)

    def test_client_01(self):
        for n in {1, 5, 10, 20, 100}:
            with self.subTest(tasksnum=n):
                client = self.create_client(max_concurrency=10)

                def worker():
                    self.assertEqual(client.query_single("SELECT 1"), 1)

                tasks = [threading.Thread(target=worker) for _ in range(n)]
                for task in tasks:
                    task.start()
                for task in tasks:
                    task.join()
                client.close()

    def test_client_02(self):
        for n in {1, 3, 5, 10, 20, 100}:
            with self.subTest(tasksnum=n):
                with self.create_client(max_concurrency=5) as client:

                    def worker():
                        self.assertEqual(client.query_single("SELECT 1"), 1)

                    tasks = [threading.Thread(target=worker) for _ in range(n)]
                    for task in tasks:
                        task.start()
                    for task in tasks:
                        task.join()

    def test_client_05(self):
        for n in {1, 3, 5, 10, 20, 100}:
            with self.subTest(tasksnum=n):
                client = self.create_client(max_concurrency=10)

                def worker():
                    self.assertEqual(client.query('SELECT 1'), [1])
                    self.assertEqual(client.query_single('SELECT 1'), 1)
                    self.assertEqual(client.query_json('SELECT 1'), '[1]')
                    self.assertEqual(client.query_single_json('SELECT 1'), '1')

                tasks = [threading.Thread(target=worker) for _ in range(n)]
                for task in tasks:
                    task.start()
                for task in tasks:
                    task.join()
                client.close()

    def test_client_transaction(self):
        client = self.create_client(max_concurrency=1)

        for tx in client.transaction():
            with tx:
                self.assertEqual(tx.query_single("SELECT 7*8"), 56)

        client.close()

    def test_client_options(self):
        client = self.create_client(max_concurrency=1)

        client.with_transaction_options(
            edgedb.TransactionOptions(readonly=True))
        client.with_retry_options(
            edgedb.RetryOptions(attempts=1, backoff=edgedb.default_backoff))
        for tx in client.transaction():
            with tx:
                self.assertEqual(tx.query_single("SELECT 7*8"), 56)

        client.close()

    def test_client_init_run_until_complete(self):
        client = self.create_client()
        self.assertIsInstance(client, blocking_client.Client)
        client.close()

    def test_client_no_acquire_deadlock(self):
        with self.create_client(
            max_concurrency=1,
        ) as client:

            has_sleep = client.query_single("""
                SELECT EXISTS(
                    SELECT schema::Function FILTER .name = 'sys::_sleep'
                )
            """)
            if not has_sleep:
                self.skipTest("No sys::_sleep function")

            def sleep_and_release():
                client.execute("SELECT sys::_sleep(1)")

            task = threading.Thread(target=sleep_and_release)
            task.start()
            time.sleep(0.5)

            client.query_single("SELECT 1")
            task.join()

    def test_client_config_persistence(self):
        N = 100

        class MyConnection(blocking_client.BlockingIOConnection):
            async def raw_query(self, query_context):
                res = await super().raw_query(query_context)
                return res + 1

        def test():
            for tx in client.transaction():
                with tx:
                    self.assertEqual(tx.query_single("SELECT 1"), 2)
                    # make this test more reliable by spending more time in
                    # the transaction to hit max_concurrency sooner
                    time.sleep(random.random() / 100)

        with self.create_client(
            max_concurrency=10,
            connection_class=MyConnection,
        ) as client:

            tasks = [threading.Thread(target=test) for _ in range(N)]
            for task in tasks:
                task.start()
            for task in tasks:
                task.join()

            self.assertEqual(
                sum(
                    1
                    for ch in client._impl._holders
                    if ch._con and not ch._con.is_closed()
                ),
                10,
            )

    def test_client_connection_methods(self):
        def test_query(client, q):
            i = random.randint(0, 20)
            time.sleep(random.random() / 100)
            r = client.query("SELECT {}".format(i))
            self.assertEqual(list(r), [i])
            q.put(1)

        def test_query_single(client, q):
            i = random.randint(0, 20)
            time.sleep(random.random() / 100)
            r = client.query_single("SELECT {}".format(i))
            self.assertEqual(r, i)
            q.put(1)

        def test_execute(client, q):
            time.sleep(random.random() / 100)
            client.execute("SELECT {1, 2, 3, 4}")
            q.put(1)

        def run(N, meth):
            with self.create_client(max_concurrency=10) as client:
                q = queue.Queue()
                coros = [
                    threading.Thread(target=meth, args=(client, q))
                    for _ in range(N)
                ]
                for coro in coros:
                    coro.start()
                for coro in coros:
                    coro.join()
                res = []
                while not q.empty():
                    res.append(q.get_nowait())
                self.assertEqual(res, [1] * N)

        methods = [
            test_query,
            test_query_single,
            test_execute,
        ]

        for method in methods:
            with self.subTest(method=method.__name__):
                run(200, method)

    def test_client_handles_transaction_exit_in_gen_1(self):
        client = self.create_client(max_concurrency=1)

        def iterate():
            for tx in client.transaction():
                with tx:
                    for record in tx.query("SELECT {1, 2, 3}"):
                        yield record

        class MyException(Exception):
            pass

        with self.assertRaises(MyException):
            agen = iterate()
            try:
                for _ in agen:  # noqa
                    raise MyException()
            finally:
                agen.close()

        client.close()

    def test_client_handles_transaction_exit_in_gen_2(self):
        client = self.create_client(max_concurrency=1)

        def iterate():
            for tx in client.transaction():
                with tx:
                    for record in tx.query("SELECT {1, 2, 3}"):
                        yield record

        class MyException(Exception):
            pass

        with self.assertRaises(MyException):
            iterator = iterate()
            try:
                for _ in iterator:  # noqa
                    raise MyException()
            finally:
                iterator.close()

            del iterator

        client.close()

    def test_client_handles_gen_finalization(self):
        client = self.create_client(max_concurrency=1)

        def iterate(tx):
            for record in tx.query("SELECT {1, 2, 3}"):
                yield record

        class MyException(Exception):
            pass

        with self.assertRaises(MyException):
            for tx in client.transaction():
                with tx:
                    agen = iterate(tx)
                    try:
                        for _ in agen:  # noqa
                            raise MyException()
                    finally:
                        agen.close()

        client.close()

    def test_client_close_waits_for_release(self):
        client = self.create_client(max_concurrency=1)

        flag = threading.Event()
        conn_released = False

        def worker():
            nonlocal conn_released

            for tx in client.transaction():
                with tx:
                    tx.query("SELECT 42")
                    flag.set()
                    time.sleep(0.1)

            conn_released = True

        task = threading.Thread(target=worker)
        task.start()

        flag.wait()
        client.close()
        self.assertTrue(conn_released)
        task.join()

    def test_client_close_timeout(self):
        client = self.create_client(max_concurrency=1)

        flag = threading.Event()

        def worker():
            with self.assertRaises(errors.ClientConnectionClosedError):
                for tx in client.transaction():
                    with tx:
                        tx.query_single("SELECT 42")
                        flag.set()
                        time.sleep(0.5)

        task = threading.Thread(target=worker)
        task.start()

        with self.assertRaises(errors.InterfaceError):
            flag.wait()
            client.close(timeout=0.1)

        task.join()

    def test_client_expire_connections(self):
        client = self.create_client(max_concurrency=1)

        for tx in client.transaction():
            with tx:
                tx.query("SELECT 42")
                client._impl.expire_connections()

        self.assertIsNone(client._impl._holders[0]._con)
        client.close()

    def test_client_properties(self):
        max_concurrency = 2

        client = self.create_client(max_concurrency=max_concurrency)
        self.assertEqual(client.max_concurrency, max_concurrency)
        self.assertEqual(client.max_concurrency, max_concurrency)

        for tx in client.transaction():
            with tx:
                tx.query("SELECT 42")
                self.assertEqual(client.free_size, max_concurrency - 1)

        self.assertEqual(client.free_size, max_concurrency)

        client.close()

    def _test_connection_broken(self, executor, broken_evt):
        self.loop.call_soon_threadsafe(broken_evt.set)

        with self.assertRaises(errors.ClientConnectionError):
            executor.query_single("SELECT 123")

        self.loop.call_soon_threadsafe(broken_evt.clear)

        self.assertEqual(executor.query_single("SELECT 123"), 123)
        self.loop.call_soon_threadsafe(broken_evt.set)
        with self.assertRaises(errors.ClientConnectionError):
            executor.query_single("SELECT 123")
        self.loop.call_soon_threadsafe(broken_evt.clear)
        self.assertEqual(executor.query_single("SELECT 123"), 123)
        self.loop.call_soon_threadsafe(broken_evt.set)
        self.loop.call_soon_threadsafe(broken_evt.clear)
        self.assertEqual(executor.query_single("SELECT 123"), 123)

        tested = False
        for tx in executor.transaction():
            with tx:
                self.assertEqual(tx.query_single("SELECT 123"), 123)
                if tested:
                    break
                tested = True
                self.loop.call_soon_threadsafe(broken_evt.set)
                try:
                    tx.query_single("SELECT 123")
                except errors.ClientConnectionError:
                    self.loop.call_soon_threadsafe(broken_evt.clear)
                    raise
                else:
                    self.fail("ConnectionError not raised!")

    async def test_client_connection_broken(self):
        con_args = self.get_connect_args()
        broken = asyncio.Event()
        done = asyncio.Event()

        async def proxy(r: asyncio.StreamReader, w: asyncio.StreamWriter):
            while True:
                reader = self.loop.create_task(r.read(65536))
                waiter = self.loop.create_task(broken.wait())
                await asyncio.wait(
                    [reader, waiter],
                    return_when=asyncio.FIRST_COMPLETED,
                )
                if waiter.done():
                    reader.cancel()
                    w.close()
                    break
                else:
                    waiter.cancel()
                    data = await reader
                    if not data:
                        w.close()
                        break
                    w.write(data)

        async def cb(r: asyncio.StreamReader, w: asyncio.StreamWriter):
            ur, uw = await asyncio.open_connection(
                con_args['host'], con_args['port']
            )
            done.clear()
            task = self.loop.create_task(proxy(r, uw))
            try:
                await proxy(ur, w)
            finally:
                try:
                    await task
                finally:
                    done.set()
                    w.close()
                    uw.close()

        server = await asyncio.start_server(
            cb, '127.0.0.1', 0
        )
        port = server.sockets[0].getsockname()[1]
        client = self.create_client(
            host='127.0.0.1',
            port=port,
            max_concurrency=1,
            wait_until_available=5,
        )
        try:
            await self.loop.run_in_executor(
                None, self._test_connection_broken, client, broken
            )
        finally:
            await self.loop.run_in_executor(None, client.close, 5)
            server.close()
            await server.wait_closed()
            broken.set()
            await done.wait()

    def test_client_suggested_concurrency(self):
        conargs = self.get_connect_args().copy()
        conargs["database"] = self.get_database_name()
        conargs["timeout"] = 120

        client = edgedb.create_client(**conargs)

        self.assertEqual(client.max_concurrency, 1)

        client.ensure_connected()
        self.assertGreater(client.max_concurrency, 1)

        client.close()

        client = edgedb.create_client(**conargs, max_concurrency=5)

        self.assertEqual(client.max_concurrency, 5)

        client.ensure_connected()
        self.assertEqual(client.max_concurrency, 5)

        client.close()



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2024-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import unittest

try:
    from sqlalchemy import create_engine
    from sqlalchemy.orm import Session
except ImportError:
    NO_ORM = True
else:
    NO_ORM = False

from gel import _testbase as tb


class TestSQLAFeatures(tb.SQLATestCase):
    SCHEMA = os.path.join(
        os.path.dirname(__file__), 'dbsetup', 'features_default.esdl')

    SCHEMA_OTHER = os.path.join(
        os.path.dirname(__file__), 'dbsetup', 'features_other.esdl')

    SCHEMA_OTHER_NESTED = os.path.join(
        os.path.dirname(__file__), 'dbsetup', 'features_other_nested.esdl')

    SETUP = os.path.join(os.path.dirname(__file__), 'dbsetup',
                         'features.edgeql')

    MODEL_PACKAGE = 'fmodels'

    @classmethod
    def setUpClass(cls):
        if NO_ORM:
            raise unittest.SkipTest("sqlalchemy is not installed")

        super().setUpClass()
        cls.engine = create_engine(cls.get_dsn_for_sqla())
        cls.sess = Session(cls.engine, autobegin=False)

        from fmodels import default, other
        from fmodels.other import nested
        cls.sm = default
        cls.sm_o = other
        cls.sm_on = nested

    def setUp(self):
        super().setUp()

        if self.client.query_required_single('''
            select sys::get_version().major < 6
        '''):
            self.skipTest("Test needs SQL DML queries")

        self.sess.begin()

    def tearDown(self):
        super().tearDown()
        self.sess.rollback()

    def test_sqla_linkprops_01(self):
        val = self.sess.query(self.sm.HasLinkPropsA).one()
        self.assertEqual(val.child.target.num, 0)
        self.assertEqual(val.child.a, 'single')

    def test_sqla_linkprops_02(self):
        val = self.sess.query(self.sm.HasLinkPropsA).one()
        self.assertEqual(val.child.target.num, 0)
        self.assertEqual(val.child.a, 'single')

        # replace the single child with a different one
        ch = self.sess.query(self.sm.Child).filter_by(num=1).one()
        val.child = self.sm.HasLinkPropsA_child_link(a='replaced', target=ch)
        self.sess.flush()

        val = self.sess.query(self.sm.HasLinkPropsA).one()
        self.assertEqual(val.child.target.num, 1)
        self.assertEqual(val.child.a, 'replaced')

        # make sure there's only one link object still
        vals = self.sess.query(self.sm.HasLinkPropsA_child_link).all()
        self.assertEqual(
            [(val.a, val.target.num) for val in vals],
            [('replaced', 1)]
        )

    def test_sqla_linkprops_03(self):
        val = self.sess.query(self.sm.HasLinkPropsA).one()
        self.assertEqual(val.child.target.num, 0)
        self.assertEqual(val.child.a, 'single')

        # delete the child object
        val = self.sess.query(self.sm.Child).filter_by(num=0).one()
        self.sess.delete(val)
        self.sess.flush()

        val = self.sess.query(self.sm.HasLinkPropsA).one()
        self.assertEqual(val.child, None)

        # make sure the link object is removed
        vals = self.sess.query(self.sm.HasLinkPropsA_child_link).all()
        self.assertEqual(vals, [])

    def test_sqla_linkprops_04(self):
        val = self.sess.query(self.sm.HasLinkPropsB).one()
        self.assertEqual(
            {(c.b, c.target.num) for c in val.children},
            {('hello', 0), ('world', 1)},
        )

    def test_sqla_linkprops_05(self):
        val = self.sess.query(self.sm.HasLinkPropsB).one()
        self.assertEqual(
            {(c.b, c.target.num) for c in val.children},
            {('hello', 0), ('world', 1)},
        )

        # Remove one of the children
        for t in list(val.children):
            if t.b != 'hello':
                val.children.remove(t)
        self.sess.flush()

        val = self.sess.query(self.sm.HasLinkPropsB).one()
        self.assertEqual(
            {(c.b, c.target.num) for c in val.children},
            {('hello', 0)},
        )

    def test_sqla_linkprops_06(self):
        val = self.sess.query(self.sm.HasLinkPropsB).one()
        self.assertEqual(
            {(c.b, c.target.num) for c in val.children},
            {('hello', 0), ('world', 1)},
        )

        # Remove one of the children
        val = self.sess.query(self.sm.Child).filter_by(num=0).one()
        self.sess.delete(val)
        self.sess.flush()

        val = self.sess.query(self.sm.HasLinkPropsB).one()
        self.assertEqual(
            {(c.b, c.target.num) for c in val.children},
            {('world', 1)},
        )

    def test_sqla_module_01(self):
        vals = self.sess.query(self.sm_o.Branch).all()
        self.assertEqual(
            {(r.val, tuple(sorted(lf.num for lf in r.leaves))) for r in vals},
            {
                ('big', (20, 30)),
                ('small', (10,)),
            },
        )

        vals = self.sess.query(self.sm_on.Leaf).all()
        self.assertEqual(
            {r.num for r in vals},
            {10, 20, 30},
        )

        vals = self.sess.query(self.sm.Theme).all()
        self.assertEqual(
            {
                (r.color, r.branch.note, r.branch.target.val)
                for r in vals
            },
            {
                ('green', 'fresh', 'big'),
                ('orange', 'fall', 'big'),
            },
        )

    def test_sqla_module_02(self):
        val = self.sess.query(self.sm.Theme).filter_by(color='orange').one()
        self.assertEqual(
            (val.color, val.branch.note, val.branch.target.val),
            ('orange', 'fall', 'big'),
        )

        # swap the branch for 'small'
        br = self.sess.query(self.sm_o.Branch).filter_by(val='small').one()
        # can't update link tables (Gel limitation), so we will create a new
        # one
        val.branch = self.sm.Theme_branch_link(
            note='swapped', target=br)
        self.sess.add(val)
        self.sess.flush()

        vals = self.sess.query(self.sm.Theme).all()
        self.assertEqual(
            {
                (r.color, r.branch.note, r.branch.target.val)
                for r in vals
            },
            {
                ('green', 'fresh', 'big'),
                ('orange', 'swapped', 'small'),
            },
        )

    def test_sqla_bklink_01(self):
        # test backlink name collisions
        foo = self.sess.query(self.sm.Foo).filter_by(name='foo').one()
        oof = self.sess.query(self.sm.Foo).filter_by(name='oof').one()

        # only one link from Bar 123 to foo
        self.assertEqual(
            [obj.n for obj in foo._foo_Bar],
            [123],
        )
        # only one link from Who 456 to oof
        self.assertEqual(
            [obj.x for obj in oof._foo_Who],
            [456],
        )

        # foo is linked via `many_foo` from both Bar and Who
        self.assertEqual(
            [obj.n for obj in foo._many_foo_Bar],
            [123],
        )
        self.assertEqual(
            [(obj.note, obj.source.x) for obj in foo._many_foo_Who],
            [('just one', 456)],
        )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import (
    NoReturn,
    Tuple,
)

import os
import os.path
import pathlib
import platform
import shutil
import ssl
import stat
import subprocess
import sys
import tempfile
import time
import urllib.request


PACKAGE_URL_PREFIX = "https://packages.edgedb.com/dist"
STRONG_CIPHERSUITES = ":".join([
    "TLS_AES_128_GCM_SHA256",
    "TLS_CHACHA20_POLY1305_SHA256",
    "TLS_AES_256_GCM_SHA384",
    "ECDHE-ECDSA-AES128-GCM-SHA256",
    "ECDHE-RSA-AES128-GCM-SHA256",
    "ECDHE-ECDSA-CHACHA20-POLY1305",
    "ECDHE-RSA-CHACHA20-POLY1305",
    "ECDHE-ECDSA-AES256-GCM-SHA384",
    "ECDHE-RSA-AES256-GCM-SHA384",
])


def _die(msg: str) -> NoReturn:
    print(f"error: {msg}", file=sys.stderr)
    sys.exit(1)


def _warn(msg: str) -> NoReturn:
    print(f"warning: {msg}", file=sys.stderr)


def _run_cli(path: str) -> NoReturn:
    cmd = [path] + sys.argv[1:]
    if os.name == "nt":
        result = subprocess.run(cmd)
        sys.exit(result.returncode)
    else:
        os.execv(path, cmd)


def _real_mac_machine(machine: str) -> str:
    import ctypes
    import ctypes.util

    def _sysctl(libc: ctypes.CDLL, name: str) -> str:
        size = ctypes.c_uint(0)
        libc.sysctlbyname(name, None, ctypes.byref(size), None, 0)
        buf = ctypes.create_string_buffer(size.value)
        libc.sysctlbyname(name, buf, ctypes.byref(size), None, 0)
        return buf.value

    libc_path = ctypes.util.find_library("c")
    if not libc_path:
        _die("could not find the C library")
    libc = ctypes.CDLL(libc_path)
    if machine == "i386":
        # check for 32-bit emulation on a 64-bit x86 machine
        if _sysctl(libc, "hw.optional.x86_64") == "1":
            machine = "x86_64"
    elif machine == "x86_64":
        # check for Rosetta
        if _sysctl(libc, "sysctl.proc_translated") == "1":
            machine = "aarch64"

    return machine


def _platform() -> Tuple[str, str]:
    uname = platform.uname()
    uname_sys = uname.system
    machine = uname.machine.lower()
    if (
        uname_sys == "Windows"
        or uname_sys.startswith("CYGWIN_NT")
        or uname_sys.startswith("MINGW64_NT")
        or uname_sys.startswith("MSYS_NT")
    ):
        os = "Windows"
    elif uname_sys == "Darwin":
        if machine == "i386" or machine == "x86_64":
            machine = _real_mac_machine(machine)
        os = "Darwin"
    elif uname_sys == "Linux":
        os = "Linux"
    else:
        _die(f"unsupported OS: {uname_sys}")

    if machine in ("x86-64", "x64", "amd64"):
        machine = "x86_64"
    elif machine == "arm64":
        machine = "aarch64"

    if machine not in ("x86_64", "aarch64") or (
        machine == "aarch64" and os not in ("Darwin", "Linux")
    ):
        _die(f"unsupported hardware architecture: {machine}")

    return os, machine


def _download(url: str, dest: pathlib.Path) -> None:
    if not url.lower().startswith("https://"):
        _die(f"unexpected insecure URL: {url}")

    # Create an SSL context with certificate verification enabled
    ssl_context = ssl.create_default_context()
    ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
    ssl_context.set_ciphers(STRONG_CIPHERSUITES)

    try:
        # Open the URL with the SSL context
        with urllib.request.urlopen(url, context=ssl_context) as response:
            final_url = response.geturl()
            if not final_url.lower().startswith("https://"):
                _die("redirected to a non-HTTPS URL, download aborted.")

            if response.status != 200:
                raise RuntimeError(f"{response.status}")

            spinner_symbols = ['|', '/', '-', '\\']
            msg = "downloading Gel CLI"
            print(f"{msg}", end="\r")
            start = time.monotonic()

            with open(str(dest), mode="wb") as file:
                i = 0
                while True:
                    chunk = response.read(524288)
                    if not chunk:
                        break
                    file.write(chunk)
                    now = time.monotonic()
                    if now - start > 0.2:
                        print(f"\r{msg} {spinner_symbols[i]}", end="\r")
                        start = now
                        i = (i + 1) % len(spinner_symbols)

            # clear
            print(f"{' ' * (len(msg) + 2)}", end="\r")

    except Exception as e:
        _die(f"could not download Gel CLI: {e}")


def _get_binary_cache_dir(os_name) -> pathlib.Path:
    home = pathlib.Path.home()
    if os_name == 'Windows':
        localappdata = os.environ.get('LOCALAPPDATA', '')
        if localappdata:
            base_cache_dir = pathlib.Path(localappdata)
        else:
            base_cache_dir = home / 'AppData' / 'Local'
    elif os_name == 'Linux':
        xdg_cache_home = os.environ.get('XDG_CACHE_HOME', '')
        if xdg_cache_home:
            base_cache_dir = pathlib.Path(xdg_cache_home)
        else:
            base_cache_dir = home / '.cache'
    elif os_name == 'Darwin':
        base_cache_dir = home / 'Library' / 'Caches'
    else:
        _die(f"unsupported OS: {os_name}")

    cache_dir = base_cache_dir / "gel" / "bin"
    try:
        cache_dir.mkdir(parents=True, exist_ok=True)
    except OSError as e:
        _warn(f"could not create {cache_dir}: {e}")

        try:
            cache_dir = pathlib.Path(tempfile.mkdtemp(prefix="gel"))
        except Exception as e:
            _die(f"could not create temporary directory: {e}")

    return cache_dir


def _get_mountpoint(path: pathlib.Path) -> pathlib.Path:
    path = path.resolve()
    if os.path.ismount(str(path)):
        return path
    else:
        for p in path.parents:
            if os.path.ismount(str(p)):
                return p

        return p


def _install_cli(os_name: str, arch: str, path: pathlib.Path) -> str:
    triple = f"{arch}"
    ext = ""
    if os_name == "Windows":
        triple += "-pc-windows-msvc"
        ext = ".exe"
    elif os_name == "Darwin":
        triple += "-apple-darwin"
    elif os_name == "Linux":
        triple += "-unknown-linux-musl"
    else:
        _die(f"unexpected OS: {os}")

    url = f"{PACKAGE_URL_PREFIX}/{triple}/edgedb-cli{ext}"

    if path.exists() and not path.is_file():
        _die(f"{path} exists but is not a regular file, "
             f"please remove it and try again")

    _download(url, path)

    try:
        path.chmod(
            stat.S_IRWXU
            | stat.S_IRGRP | stat.S_IXGRP
            | stat.S_IROTH | stat.S_IXOTH,
        )
    except OSError as e:
        _die(f"could not max {path!r} executable: {e}")

    if not os.access(str(path), os.X_OK):
        _die(
            f"cannot execute {path!r} "
            f"(likely because {_get_mountpoint(path)} is mounted as noexec)"
        )


def main() -> NoReturn:
    dev_cli = shutil.which("gel-dev")
    if dev_cli:
        path = pathlib.Path(dev_cli)
    else:
        os, arch = _platform()
        cache_dir = _get_binary_cache_dir(os)
        path = cache_dir / "gel"
        if not path.exists():
            _install_cli(os, arch, path)

    _run_cli(path)


if __name__ == "__main__":
    main()



# Auto-generated shim
TYPE_CHECKING = False
if TYPE_CHECKING:
    from gel.ai import *  # noqa
    create_ai = create_rag_client  # noqa
    EdgeDBAI = RAGClient  # noqa
    create_async_ai = create_async_rag_client  # noqa
    AsyncEdgeDBAI = AsyncRAGClient  # noqa
    AIOptions = RAGOptions  # noqa
import gel.ai as _mod
import sys as _sys
_cur = _sys.modules['edgedb.ai']
for _k in vars(_mod):
    if not _k.startswith('__') or _k in ('__all__', '__doc__'):
        setattr(_cur, _k, getattr(_mod, _k))
_cur.create_ai = _mod.create_rag_client
_cur.EdgeDBAI = _mod.RAGClient
_cur.create_async_ai = _mod.create_async_rag_client
_cur.AsyncEdgeDBAI = _mod.AsyncRAGClient
_cur.AIOptions = _mod.RAGOptions
if hasattr(_cur, '__all__'):
    _cur.__all__ = _cur.__all__ + [
        'create_ai',
        'EdgeDBAI',
        'create_async_ai',
        'AsyncEdgeDBAI',
        'AIOptions',
    ]
del _cur
del _sys
del _mod
del _k



import json
import re
import collections
import textwrap
import warnings


INTRO_QUERY = '''
with module schema
select ObjectType {
    name,
    links: {
        name,
        readonly,
        required,
        cardinality,
        exclusive := exists (
            select .constraints
            filter .name = 'std::exclusive'
        ),
        target: {name},

        properties: {
            name,
            readonly,
            required,
            cardinality,
            exclusive := exists (
                select .constraints
                filter .name = 'std::exclusive'
            ),
            target: {name},
        },
    } filter .name != '__type__' and not exists .expr,
    properties: {
        name,
        readonly,
        required,
        cardinality,
        exclusive := exists (
            select .constraints
            filter .name = 'std::exclusive'
        ),
        target: {name},
    } filter not exists .expr,
    backlinks := <array<str>>[],
}
filter
    not .builtin
    and
    not .internal
    and
    not re_test('^(std|cfg|sys|schema)::', .name);
'''

MODULE_QUERY = '''
with
    module schema,
    m := (select `Module` filter not .builtin)
select m.name;
'''

CLEAN_NAME = re.compile(r'^[A-Za-z_][A-Za-z0-9_]*$')


class GelORMWarning(Warning):
    pass


def get_sql_name(name):
    # Just remove the module name
    name = name.rsplit('::', 1)[-1]

    return name


def get_mod_and_name(name):
    # Assume the names are already validated to be properly formed
    # alphanumeric identifiers that may be prefixed by a module. If the module
    # is present assume it is safe to drop it (currently only defualt module
    # is allowed).

    # Split on module separator. Potentially if we ever handle more unusual
    # names, there may be more processing done.
    return name.rsplit('::', 1)


def valid_name(name):
    # Just remove module separators and check the rest
    name = name.replace('::', '')
    if not CLEAN_NAME.fullmatch(name):
        warnings.warn(
            f'Skipping {name!r}: non-alphanumeric names are not supported',
            GelORMWarning,
        )
        return False
    return True


def get_schema_json(client):
    types = json.loads(client.query_json(INTRO_QUERY))
    modules = json.loads(client.query_json(MODULE_QUERY))

    return _process_links(types, modules)


async def async_get_schema_json(client):
    types = json.loads(await client.query_json(INTRO_QUERY))
    modules = json.loads(client.query_json(MODULE_QUERY))

    return _process_links(types, modules)


def _skip_invalid_names(spec_list, recurse_into=None):
    valid = []
    for spec in spec_list:
        # skip invalid names
        if valid_name(spec['name']):
            if recurse_into is not None:
                for fname in recurse_into:
                    if fname not in spec:
                        continue
                    spec[fname] = _skip_invalid_names(
                        spec[fname], recurse_into)

            valid.append(spec)

    return valid


def _process_links(types, modules):
    # Figure out all the backlinks, link tables, and links with link
    # properties that require their own intermediate objects.
    type_map = {}
    link_tables = []
    link_objects = []
    prop_objects = []

    # All the names of types, props and links are valid beyond this point.
    types = _skip_invalid_names(types, ['properties', 'links'])
    for spec in types:
        type_map[spec['name']] = spec
        spec['backlink_renames'] = {}

    for spec in types:
        mod = spec["name"].rsplit('::', 1)[0]
        sql_source = get_sql_name(spec["name"])

        for prop in spec['properties']:
            name = prop['name']
            exclusive = prop['exclusive']
            cardinality = prop['cardinality']
            sql_name = get_sql_name(name)

            if cardinality == 'Many':
                # Multi property will make its own "link table". But since it
                # doesn't link to any other object the link table itself must
                # be reflected as an object.
                pobj = {
                    'module': mod,
                    'name': f'{sql_source}_{sql_name}_prop',
                    'table': f'{sql_source}.{sql_name}',
                    'links': [{
                        'name': 'source',
                        'required': True,
                        'cardinality': 'One' if exclusive else 'Many',
                        'exclusive': cardinality == 'One',
                        'target': {'name': spec['name']},
                        'has_link_object': False,
                    }],
                    'properties': [{
                        'name': 'target',
                        'required': True,
                        'cardinality': 'One',
                        'exclusive': False,
                        'target': prop['target'],
                        'has_link_object': False,
                    }],
                }
                prop_objects.append(pobj)

        for link in spec['links']:
            if link['name'] != '__type__':
                name = link['name']
                target = link['target']['name']
                cardinality = link['cardinality']
                exclusive = link['exclusive']
                sql_name = get_sql_name(name)

                objtype = type_map[target]
                objtype['backlinks'].append({
                    # naming scheme mimics .<link[is Type]
                    'name': f'_{sql_name}_{sql_source}',
                    'fwname': sql_name,
                    # flip cardinality and exclusivity
                    'cardinality': 'One' if exclusive else 'Many',
                    'exclusive': cardinality == 'One',
                    'target': {'name': spec['name']},
                    'has_link_object': False,
                })

                link['has_link_object'] = False
                # Any link with properties should become its own intermediate
                # object, since ORMs generally don't have a special convenient
                # way of exposing this as just a link table.
                if len(link['properties']) > 2:
                    # more than just 'source' and 'target' properties
                    lobj = {
                        'module': mod,
                        'name': f'{sql_source}_{sql_name}_link',
                        'table': f'{sql_source}.{sql_name}',
                        'links': [],
                        'properties': [],
                    }
                    for prop in link['properties']:
                        if prop['name'] in {'source', 'target'}:
                            lobj['links'].append(prop)
                        else:
                            lobj['properties'].append(prop)

                    link_objects.append(lobj)
                    link['has_link_object'] = True
                    objtype['backlinks'][-1]['has_link_object'] = True

                elif cardinality == 'Many':
                    # Add a link table for One-to-Many and Many-to-Many
                    link_tables.append({
                        'module': mod,
                        'name': f'{sql_source}_{sql_name}_table',
                        'table': f'{sql_source}.{sql_name}',
                        'source': spec["name"],
                        'target': target,
                    })

    return {
        'modules': modules,
        'object_types': types,
        'link_tables': link_tables,
        'link_objects': link_objects,
        'prop_objects': prop_objects,
    }


class FilePrinter(object):
    INDENT = ' ' * 4

    def __init__(self):
        # set the output to be stdout by default, but this is generally
        # expected to be overridden
        self.out = None
        self._indent_level = 0

    def indent(self):
        self._indent_level += 1

    def dedent(self):
        if self._indent_level > 0:
            self._indent_level -= 1

    def reset_indent(self):
        self._indent_level = 0

    def write(self, text=''):
        print(
            textwrap.indent(text, prefix=self.INDENT * self._indent_level),
            file=self.out,
        )



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2021-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import enum


class Capability(enum.IntFlag):

    NONE              = 0         # noqa
    MODIFICATIONS     = 1 << 0    # noqa
    SESSION_CONFIG    = 1 << 1    # noqa
    TRANSACTION       = 1 << 2    # noqa
    DDL               = 1 << 3    # noqa
    PERSISTENT_CONFIG = 1 << 4    # noqa

    ALL               = 0xFFFF_FFFF_FFFF_FFFF                 # noqa
    EXECUTE           = ALL & ~TRANSACTION & ~SESSION_CONFIG  # noqa
    LEGACY_EXECUTE    = ALL & ~TRANSACTION                    # noqa


class CompilationFlag(enum.IntFlag):

    INJECT_OUTPUT_TYPE_IDS   = 1 << 0    # noqa
    INJECT_OUTPUT_TYPE_NAMES = 1 << 1    # noqa
    INJECT_OUTPUT_OBJECT_IDS = 1 << 2    # noqa


class Cardinality(enum.Enum):
    # Cardinality isn't applicable for the query:
    # * the query is a command like CONFIGURE that
    #   does not return any data;
    # * the query is composed of multiple queries.
    NO_RESULT = 0x6e

    # Cardinality is 1 or 0
    AT_MOST_ONE = 0x6f

    # Cardinality is 1
    ONE = 0x41

    # Cardinality is >= 0
    MANY = 0x6d

    # Cardinality is >= 1
    AT_LEAST_ONE = 0x4d

    def is_single(self) -> bool:
        return self in {Cardinality.AT_MOST_ONE, Cardinality.ONE}

    def is_multi(self) -> bool:
        return self in {Cardinality.AT_LEAST_ONE, Cardinality.MANY}


class ElementKind(enum.Enum):

    LINK            = 1 # noqa
    PROPERTY        = 2 # noqa
    LINK_PROPERTY   = 3 # noqa



import abc
import enum
import logging
import random
import typing
import sys
from collections import namedtuple

from . import errors


logger = logging.getLogger('gel')


_RetryRule = namedtuple("_RetryRule", ["attempts", "backoff"])
TAG_NAME = "tag"


def default_backoff(attempt):
    return (2 ** attempt) * 0.1 + random.randrange(100) * 0.001


WarningHandler = typing.Callable[
    [typing.Tuple[errors.EdgeDBError, ...], typing.Any],
    typing.Any,
]


def raise_warnings(warnings, res):
    if (
        len(warnings) > 1
        and sys.version_info >= (3, 11)
    ):
        raise ExceptionGroup(  # noqa
            "Query produced warnings", warnings
        )
    else:
        raise warnings[0]


def log_warnings(warnings, res):
    for w in warnings:
        logger.warning("EdgeDB warning: %s", str(w))
    return res


class RetryCondition:
    """Specific condition to retry on for fine-grained control"""
    TransactionConflict = enum.auto()
    NetworkError = enum.auto()


class IsolationLevel:
    """Isolation level for transaction"""
    Serializable = "SERIALIZABLE"
    RepeatableRead = "REPEATABLE READ"


class RetryOptions:
    """An immutable class that contains rules for `transaction()`"""
    __slots__ = ['_default', '_overrides']

    def __init__(self, attempts: int, backoff=default_backoff):
        self._default = _RetryRule(attempts, backoff)
        self._overrides = None

    def with_rule(self, condition, attempts=None, backoff=None):
        default = self._default
        overrides = self._overrides
        if overrides is None:
            overrides = {}
        else:
            overrides = overrides.copy()
        overrides[condition] = _RetryRule(
            default.attempts if attempts is None else attempts,
            default.backoff if backoff is None else backoff,
        )
        result = RetryOptions.__new__(RetryOptions)
        result._default = default
        result._overrides = overrides
        return result

    @classmethod
    def defaults(cls):
        return cls(
            attempts=3,
            backoff=default_backoff,
        )

    def get_rule_for_exception(self, exception):
        default = self._default
        overrides = self._overrides
        res = default
        if overrides:
            if isinstance(exception, errors.TransactionConflictError):
                res = overrides.get(RetryCondition.TransactionConflict, res)
            elif isinstance(exception, errors.ClientError):
                res = overrides.get(RetryCondition.NetworkError, res)
        return res


class TransactionOptions:
    """Options for `transaction()`"""
    __slots__ = ['_isolation', '_readonly', '_deferrable']

    def __init__(
        self,
        isolation: IsolationLevel=IsolationLevel.Serializable,
        readonly: bool = False,
        deferrable: bool = False,
    ):
        self._isolation = isolation
        self._readonly = readonly
        self._deferrable = deferrable

    @classmethod
    def defaults(cls):
        return cls()

    def start_transaction_query(self):
        isolation = str(self._isolation)
        if self._readonly:
            mode = 'READ ONLY'
        else:
            mode = 'READ WRITE'

        if self._deferrable:
            defer = 'DEFERRABLE'
        else:
            defer = 'NOT DEFERRABLE'

        return f'START TRANSACTION ISOLATION {isolation}, {mode}, {defer};'

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} '
            f'isolation:{self._isolation}, '
            f'readonly:{self._readonly}, '
            f'deferrable:{self._deferrable}>'
        )


class State:
    __slots__ = ['_module', '_aliases', '_config', '_globals']

    def __init__(
        self,
        default_module: typing.Optional[str] = None,
        module_aliases: typing.Mapping[str, str] = None,
        config: typing.Mapping[str, typing.Any] = None,
        globals_: typing.Mapping[str, typing.Any] = None,
    ):
        self._module = default_module
        self._aliases = {} if module_aliases is None else dict(module_aliases)
        self._config = {} if config is None else dict(config)
        self._globals = (
            {} if globals_ is None else self.with_globals(globals_)._globals
        )

    @classmethod
    def _new(cls, default_module, module_aliases, config, globals_):
        rv = cls.__new__(cls)
        rv._module = default_module
        rv._aliases = module_aliases
        rv._config = config
        rv._globals = globals_
        return rv

    @classmethod
    def defaults(cls):
        return cls()

    def with_default_module(self, module: typing.Optional[str] = None):
        return self._new(
            default_module=module,
            module_aliases=self._aliases,
            config=self._config,
            globals_=self._globals,
        )

    def with_module_aliases(self, *args, **aliases):
        if len(args) > 1:
            raise errors.InvalidArgumentError(
                "with_module_aliases() takes from 0 to 1 positional arguments "
                "but {} were given".format(len(args))
            )
        aliases_dict = args[0] if args else {}
        aliases_dict.update(aliases)
        new_aliases = self._aliases.copy()
        new_aliases.update(aliases_dict)
        return self._new(
            default_module=self._module,
            module_aliases=new_aliases,
            config=self._config,
            globals_=self._globals,
        )

    def with_config(self, *args, **config):
        if len(args) > 1:
            raise errors.InvalidArgumentError(
                "with_config() takes from 0 to 1 positional arguments "
                "but {} were given".format(len(args))
            )
        config_dict = args[0] if args else {}
        config_dict.update(config)
        new_config = self._config.copy()
        new_config.update(config_dict)
        return self._new(
            default_module=self._module,
            module_aliases=self._aliases,
            config=new_config,
            globals_=self._globals,
        )

    def resolve(self, name: str) -> str:
        parts = name.split("::", 1)
        if len(parts) == 1:
            return f"{self._module or 'default'}::{name}"
        elif len(parts) == 2:
            mod, name = parts
            mod = self._aliases.get(mod, mod)
            return f"{mod}::{name}"
        else:
            raise AssertionError('broken split')

    def with_globals(self, *args, **globals_):
        if len(args) > 1:
            raise errors.InvalidArgumentError(
                "with_globals() takes from 0 to 1 positional arguments "
                "but {} were given".format(len(args))
            )
        new_globals = self._globals.copy()
        if args:
            for k, v in args[0].items():
                new_globals[self.resolve(k)] = v
        for k, v in globals_.items():
            new_globals[self.resolve(k)] = v
        return self._new(
            default_module=self._module,
            module_aliases=self._aliases,
            config=self._config,
            globals_=new_globals,
        )

    def without_module_aliases(self, *aliases):
        if not aliases:
            new_aliases = {}
        else:
            new_aliases = self._aliases.copy()
            for alias in aliases:
                new_aliases.pop(alias, None)
        return self._new(
            default_module=self._module,
            module_aliases=new_aliases,
            config=self._config,
            globals_=self._globals,
        )

    def without_config(self, *config_names):
        if not config_names:
            new_config = {}
        else:
            new_config = self._config.copy()
            for name in config_names:
                new_config.pop(name, None)
        return self._new(
            default_module=self._module,
            module_aliases=self._aliases,
            config=new_config,
            globals_=self._globals,
        )

    def without_globals(self, *global_names):
        if not global_names:
            new_globals = {}
        else:
            new_globals = self._globals.copy()
            for name in global_names:
                new_globals.pop(self.resolve(name), None)
        return self._new(
            default_module=self._module,
            module_aliases=self._aliases,
            config=self._config,
            globals_=new_globals,
        )

    def as_dict(self):
        rv = {}
        if self._module is not None:
            rv["module"] = self._module
        if self._aliases:
            rv["aliases"] = list(self._aliases.items())
        if self._config:
            rv["config"] = self._config
        if self._globals:
            rv["globals"] = self._globals
        return rv


class _OptionsMixin:
    def __init__(self, *args, **kwargs):
        self._options = _Options.defaults()
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def _shallow_clone(self):
        pass

    def with_transaction_options(self, options: TransactionOptions = None):
        """Returns object with adjusted options for future transactions.

        :param options TransactionOptions:
            Object that encapsulates transaction options.

        This method returns a "shallow copy" of the current object
        with modified transaction options.

        Both ``self`` and returned object can be used after, but when using
        them transaction options applied will be different.

        Transaction options are used by the ``transaction`` method.
        """
        result = self._shallow_clone()
        result._options = self._options.with_transaction_options(options)
        return result

    def with_retry_options(self, options: RetryOptions=None):
        """Returns object with adjusted options for future retrying
        transactions.

        :param options RetryOptions:
            Object that encapsulates retry options.

        This method returns a "shallow copy" of the current object
        with modified retry options.

        Both ``self`` and returned object can be used after, but when using
        them retry options applied will be different.
        """

        result = self._shallow_clone()
        result._options = self._options.with_retry_options(options)
        return result

    def with_warning_handler(self, warning_handler: WarningHandler=None):
        """Returns object with adjusted options for handling warnings.

        :param warning_handler WarningHandler:
            Function for handling warnings. It is passed a tuple of warnings
            and the query result and returns a potentially updated query
            result.

        This method returns a "shallow copy" of the current object
        with modified retry options.

        Both ``self`` and returned object can be used after, but when using
        them retry options applied will be different.
        """

        result = self._shallow_clone()
        result._options = self._options.with_warning_handler(warning_handler)
        return result

    def with_state(self, state: State):
        result = self._shallow_clone()
        result._options = self._options.with_state(state)
        return result

    def with_default_module(self, module: typing.Optional[str] = None):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_default_module(module)
        )
        return result

    def with_module_aliases(self, *args, **aliases):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_module_aliases(*args, **aliases)
        )
        return result

    def with_config(self, *args, **config):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_config(*args, **config)
        )
        return result

    def with_globals(self, *args, **globals_):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_globals(*args, **globals_)
        )
        return result

    def without_module_aliases(self, *aliases):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.without_module_aliases(*aliases)
        )
        return result

    def without_config(self, *config_names):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.without_config(*config_names)
        )
        return result

    def without_globals(self, *global_names):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.without_globals(*global_names)
        )
        return result

    def with_query_tag(self, tag: str):
        for prefix in ["edgedb/", "gel/"]:
            if tag.startswith(prefix):
                raise errors.InvalidArgumentError(f"reserved tag: {prefix}*")
        if len(tag) > 128:
            raise errors.InvalidArgumentError(
                "tag too long (> 128 characters)"
            )

        result = self._shallow_clone()
        result._options = self._options.with_annotations(
            self._options.annotations | {TAG_NAME: tag}
        )
        return result

    def without_query_tag(self):
        result = self._shallow_clone()
        annotations = self._options.annotations.copy()
        annotations.pop(TAG_NAME, None)
        result._options = self._options.with_annotations(annotations)
        return result


class _Options:
    """Internal class for storing connection options"""

    __slots__ = [
        '_retry_options', '_transaction_options', '_state',
        '_warning_handler', '_annotations'
    ]

    def __init__(
        self,
        retry_options: RetryOptions,
        transaction_options: TransactionOptions,
        state: State,
        warning_handler: WarningHandler,
        annotations: typing.Dict[str, str],
    ):
        self._retry_options = retry_options
        self._transaction_options = transaction_options
        self._state = state
        self._warning_handler = warning_handler
        self._annotations = annotations

    @property
    def retry_options(self):
        return self._retry_options

    @property
    def transaction_options(self):
        return self._transaction_options

    @property
    def state(self):
        return self._state

    @property
    def warning_handler(self):
        return self._warning_handler

    @property
    def annotations(self):
        return self._annotations

    def with_retry_options(self, options: RetryOptions):
        return _Options(
            options,
            self._transaction_options,
            self._state,
            self._warning_handler,
            self._annotations,
        )

    def with_transaction_options(self, options: TransactionOptions):
        return _Options(
            self._retry_options,
            options,
            self._state,
            self._warning_handler,
            self._annotations,
        )

    def with_state(self, state: State):
        return _Options(
            self._retry_options,
            self._transaction_options,
            state,
            self._warning_handler,
            self._annotations,
        )

    def with_warning_handler(self, warning_handler: WarningHandler):
        return _Options(
            self._retry_options,
            self._transaction_options,
            self._state,
            warning_handler,
            self._annotations,
        )

    def with_annotations(self, annotations: typing.Dict[str, str]):
        return _Options(
            self._retry_options,
            self._transaction_options,
            self._state,
            self._warning_handler,
            annotations,
        )

    @classmethod
    def defaults(cls):
        return cls(
            RetryOptions.defaults(),
            TransactionOptions.defaults(),
            State.defaults(),
            log_warnings,
            {},
        )



from django.apps import AppConfig


class GelPGModel(AppConfig):
    name = "gel.orm.django.gelmodels"

    def ready(self):
        from django.db import connections, utils

        gel_compiler_module = "gel.orm.django.gelmodels.compiler"

        # Change the current compiler_module
        for c in connections:
            connections[c].ops.compiler_module = gel_compiler_module

        # Update the load_backend to use our DatabaseWrapper
        orig_load_backend = utils.load_backend

        def custom_load_backend(*args, **kwargs):
            backend = orig_load_backend(*args, **kwargs)

            class GelPGBackend:
                @staticmethod
                def DatabaseWrapper(*args2, **kwargs2):
                    connection = backend.DatabaseWrapper(*args2, **kwargs2)
                    connection.ops.compiler_module = gel_compiler_module
                    return connection

            return GelPGBackend

        utils.load_backend = custom_load_backend


#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2016-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import array
import decimal
import uuid
import datetime
from gel import describe
from gel import enums
from gel.datatypes import datatypes

from libc.string cimport memcpy
from cpython.bytes cimport PyBytes_FromStringAndSize

include "./edb_types.pxi"


include "./base.pyx"
include "./scalar.pyx"
include "./tuple.pyx"
include "./namedtuple.pyx"
include "./object.pyx"
include "./array.pyx"
include "./range.pyx"
include "./set.pyx"
include "./enum.pyx"
include "./record.pyx"


DEF CTYPE_SET = 0
DEF CTYPE_SHAPE = 1
DEF CTYPE_BASE_SCALAR = 2
DEF CTYPE_SCALAR = 3
DEF CTYPE_TUPLE = 4
DEF CTYPE_NAMEDTUPLE = 5
DEF CTYPE_ARRAY = 6
DEF CTYPE_ENUM = 7
DEF CTYPE_INPUT_SHAPE = 8
DEF CTYPE_RANGE = 9
DEF CTYPE_OBJECT = 10
DEF CTYPE_COMPOUND = 11
DEF CTYPE_MULTIRANGE = 12
DEF CTYPE_SQL_ROW = 13
DEF CTYPE_ANNO_TYPENAME = 255

DEF _CODECS_BUILD_CACHE_SIZE = 200

DEF NBASE = 10000
DEF NUMERIC_POS = 0x0000
DEF NUMERIC_NEG = 0x4000

cdef BaseCodec NULL_CODEC = NullCodec.__new__(NullCodec)
cdef BaseCodec EMPTY_TUPLE_CODEC = EmptyTupleCodec.__new__(EmptyTupleCodec)


cdef class CodecsRegistry:

    def __init__(self, *, cache_size=1000):
        self.codecs_build_cache = LRUMapping(maxsize=_CODECS_BUILD_CACHE_SIZE)
        self.codecs = LRUMapping(maxsize=cache_size)
        self.base_codec_overrides = {}

    def clear_cache(self):
        self.codecs.clear()
        self.codecs_build_cache.clear()

    def set_type_codec(self, typeid, *, encoder, decoder, format):
        if format != 'python':
            raise ValueError('"python" is the only valid format')
        if not isinstance(typeid, uuid.UUID):
            raise TypeError('typeid must be a UUID')
        basecodec = BASE_SCALAR_CODECS.get(typeid.bytes)
        if basecodec is None:
            raise ValueError(
                f'{typeid} does not correspond to any known base type')
        self.base_codec_overrides[typeid.bytes] = CodecPythonOverride.new(
            typeid.bytes,
            basecodec,
            encoder,
            decoder,
        )

    cdef BaseCodec _build_codec(self, FRBuffer *spec, list codecs_list,
                                protocol_version):
        cdef:
            uint32_t desc_len = 0
            uint8_t t
            bytes tid
            uint16_t els
            uint16_t i
            uint32_t str_len
            uint16_t pos
            int32_t dim_len
            BaseCodec res
            BaseCodec sub_codec

        if protocol_version >= (2, 0):
            desc_len = frb_get_len(spec) - 16 - 1

        t = <uint8_t>(frb_read(spec, 1)[0])
        tid = frb_read(spec, 16)[:16]

        res = self.codecs.get(tid, None)
        if res is None:
            res = self.codecs_build_cache.get(tid, None)
        if res is not None:
            # We have a codec for this "tid"; advance the buffer
            # so that we can process the next codec.
            if desc_len > 0:
                frb_read(spec, desc_len)
                return res

            if t == CTYPE_SET:
                frb_read(spec, 2)

            elif t == CTYPE_SHAPE or t == CTYPE_INPUT_SHAPE:
                els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for i in range(els):
                    frb_read(spec, 4)  # flags
                    frb_read(spec, 1)  # cardinality
                    str_len = hton.unpack_uint32(frb_read(spec, 4))
                    # read the <str> (`str_len` bytes) and <pos> (2 bytes)
                    frb_read(spec, str_len + 2)

            elif t == CTYPE_BASE_SCALAR:
                pass

            elif t == CTYPE_SCALAR:
                frb_read(spec, 2)

            elif t == CTYPE_TUPLE:
                # tuple
                els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for i in range(els):
                    frb_read(spec, 2)

            elif t == CTYPE_NAMEDTUPLE:
                els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for i in range(els):
                    str_len = hton.unpack_uint32(frb_read(spec, 4))
                    frb_read(spec, str_len + 2)

            elif t == CTYPE_ARRAY:
                frb_read(spec, 2)
                els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                if els != 1:
                    raise NotImplementedError(
                        'cannot handle arrays with more than one dimension')
                # First dimension length.
                frb_read(spec, 4)

            elif t == CTYPE_RANGE:
                frb_read(spec, 2)

            elif t == CTYPE_MULTIRANGE:
                frb_read(spec, 2)

            elif t == CTYPE_SQL_ROW:
                els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for i in range(els):
                    str_len = hton.unpack_uint32(frb_read(spec, 4))
                    frb_read(spec, str_len + 2)

            elif t == CTYPE_ENUM:
                els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for i in range(els):
                    str_len = hton.unpack_uint32(frb_read(spec, 4))
                    frb_read(spec, str_len)

            elif t == CTYPE_ANNO_TYPENAME:
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                res.type_name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                return None

            elif 0x80 & t == 0x80:
                # Ignore all other type annotations.
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                frb_read(spec, str_len)
                return None

            else:
                raise NotImplementedError(
                    f'no codec implementation for EdgeDB data class {t}')

            return res

        if t == CTYPE_SET:
            pos = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            sub_codec = <BaseCodec>codecs_list[pos]
            res = SetCodec.new(tid, sub_codec)

        elif t == CTYPE_SHAPE:
            if protocol_version >= (2, 0):
                ephemeral_free_shape = <bint>frb_read(spec, 1)[0]
                objtype_pos = <uint16_t>hton.unpack_int16(frb_read(spec, 2))

            els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            codecs = cpython.PyTuple_New(els)
            names = cpython.PyTuple_New(els)
            flags = cpython.PyTuple_New(els)
            cards = cpython.PyTuple_New(els)
            for i in range(els):
                flag = hton.unpack_uint32(frb_read(spec, 4))  # flags
                cardinality = <uint8_t>frb_read(spec, 1)[0]

                str_len = hton.unpack_uint32(frb_read(spec, 4))
                name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                pos = <uint16_t>hton.unpack_int16(frb_read(spec, 2))

                if flag & datatypes._EDGE_POINTER_IS_LINKPROP:
                    name = "@" + name
                cpython.Py_INCREF(name)
                cpython.PyTuple_SetItem(names, i, name)

                sub_codec = codecs_list[pos]
                cpython.Py_INCREF(sub_codec)
                cpython.PyTuple_SetItem(codecs, i, sub_codec)

                cpython.Py_INCREF(flag)
                cpython.PyTuple_SetItem(flags, i, flag)

                cpython.Py_INCREF(cardinality)
                cpython.PyTuple_SetItem(cards, i, cardinality)

                if protocol_version >= (2, 0):
                    source_type_pos = <uint16_t>hton.unpack_int16(
                        frb_read(spec, 2))
                    source_type = codecs_list[source_type_pos]

            res = ObjectCodec.new(
                tid, names, flags, cards, codecs, t == CTYPE_INPUT_SHAPE
            )

        elif t == CTYPE_INPUT_SHAPE:
            els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            codecs = cpython.PyTuple_New(els)
            names = cpython.PyTuple_New(els)
            flags = cpython.PyTuple_New(els)
            cards = cpython.PyTuple_New(els)
            for i in range(els):
                flag = hton.unpack_uint32(frb_read(spec, 4))  # flags
                cardinality = <uint8_t>frb_read(spec, 1)[0]

                str_len = hton.unpack_uint32(frb_read(spec, 4))
                name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                pos = <uint16_t>hton.unpack_int16(frb_read(spec, 2))

                if flag & datatypes._EDGE_POINTER_IS_LINKPROP:
                    name = "@" + name
                cpython.Py_INCREF(name)
                cpython.PyTuple_SetItem(names, i, name)

                sub_codec = codecs_list[pos]
                cpython.Py_INCREF(sub_codec)
                cpython.PyTuple_SetItem(codecs, i, sub_codec)

                cpython.Py_INCREF(flag)
                cpython.PyTuple_SetItem(flags, i, flag)

                cpython.Py_INCREF(cardinality)
                cpython.PyTuple_SetItem(cards, i, cardinality)

            res = ObjectCodec.new(
                tid, names, flags, cards, codecs, t == CTYPE_INPUT_SHAPE
            )

        elif t == CTYPE_BASE_SCALAR:
            if tid in self.base_codec_overrides:
                return self.base_codec_overrides[tid]
            else:
                res = <BaseCodec>BASE_SCALAR_CODECS[tid]

        elif t == CTYPE_SCALAR:
            if protocol_version >= (2, 0):
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                type_name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                schema_defined = <bint>frb_read(spec, 1)[0]

                ancestor_count = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                ancestors = []
                for _ in range(ancestor_count):
                    ancestor_pos = <uint16_t>hton.unpack_int16(
                        frb_read(spec, 2))
                    ancestor_codec = codecs_list[ancestor_pos]
                    if type(ancestor_codec) is not ScalarCodec:
                        raise RuntimeError(
                            f'a scalar codec expected for base scalar type, '
                            f'got {type(ancestor_codec).__name__}')
                    ancestors.append(ancestor_codec)

                if ancestor_count == 0:
                    if tid in self.base_codec_overrides:
                        res = self.base_codec_overrides[tid]
                    else:
                        res = <BaseCodec>BASE_SCALAR_CODECS[tid]
                else:
                    fundamental_codec = ancestors[-1]
                    if type(fundamental_codec) is not ScalarCodec:
                        raise RuntimeError(
                            f'a scalar codec expected for base scalar type, '
                            f'got {type(fundamental_codec).__name__}')
                    res = (<ScalarCodec>fundamental_codec).derive(tid)
                res.type_name = type_name
            else:
                fundamental_pos = <uint16_t>hton.unpack_int16(
                    frb_read(spec, 2))
                fundamental_codec = codecs_list[fundamental_pos]
                if type(fundamental_codec) is not ScalarCodec:
                    raise RuntimeError(
                        f'a scalar codec expected for base scalar type, '
                        f'got {type(fundamental_codec).__name__}')
                res = (<ScalarCodec>fundamental_codec).derive(tid)

        elif t == CTYPE_TUPLE:
            if protocol_version >= (2, 0):
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                type_name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                schema_defined = <bint>frb_read(spec, 1)[0]
                ancestor_count = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for _ in range(ancestor_count):
                    ancestor_pos = <uint16_t>hton.unpack_int16(
                        frb_read(spec, 2))
                    ancestor_codec = codecs_list[ancestor_pos]
            else:
                type_name = None
            els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            codecs = cpython.PyTuple_New(els)
            for i in range(els):
                pos = <uint16_t>hton.unpack_int16(frb_read(spec, 2))

                sub_codec = codecs_list[pos]
                cpython.Py_INCREF(sub_codec)
                cpython.PyTuple_SetItem(codecs, i, sub_codec)

            res = TupleCodec.new(tid, codecs)
            res.type_name = type_name

        elif t == CTYPE_NAMEDTUPLE:
            if protocol_version >= (2, 0):
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                type_name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                schema_defined = <bint>frb_read(spec, 1)[0]
                ancestor_count = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for _ in range(ancestor_count):
                    ancestor_pos = <uint16_t>hton.unpack_int16(
                        frb_read(spec, 2))
                    ancestor_codec = codecs_list[ancestor_pos]
            else:
                type_name = None
            els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            codecs = cpython.PyTuple_New(els)
            names = cpython.PyTuple_New(els)
            for i in range(els):
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                pos = <uint16_t>hton.unpack_int16(frb_read(spec, 2))

                cpython.Py_INCREF(name)
                cpython.PyTuple_SetItem(names, i, name)

                sub_codec = codecs_list[pos]
                cpython.Py_INCREF(sub_codec)
                cpython.PyTuple_SetItem(codecs, i, sub_codec)

            res = NamedTupleCodec.new(tid, names, codecs)
            res.type_name = type_name

        elif t == CTYPE_SQL_ROW:
            els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            codecs = cpython.PyTuple_New(els)
            names = cpython.PyTuple_New(els)
            for i in range(els):
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                pos = <uint16_t>hton.unpack_int16(frb_read(spec, 2))

                cpython.Py_INCREF(name)
                cpython.PyTuple_SetItem(names, i, name)

                sub_codec = codecs_list[pos]
                cpython.Py_INCREF(sub_codec)
                cpython.PyTuple_SetItem(codecs, i, sub_codec)

            res = RecordCodec.new(tid, names, codecs)

        elif t == CTYPE_ENUM:
            if protocol_version >= (2, 0):
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                type_name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                schema_defined = <bint>frb_read(spec, 1)[0]
                ancestor_count = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for _ in range(ancestor_count):
                    ancestor_pos = <uint16_t>hton.unpack_int16(
                        frb_read(spec, 2))
                    ancestor_codec = codecs_list[ancestor_pos]
            else:
                type_name = None
            els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            names = cpython.PyTuple_New(els)
            for i in range(els):
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)

                cpython.Py_INCREF(name)
                cpython.PyTuple_SetItem(names, i, name)

            res = EnumCodec.new(tid, names)
            res.type_name = type_name

        elif t == CTYPE_ARRAY:
            if protocol_version >= (2, 0):
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                type_name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                schema_defined = <bint>frb_read(spec, 1)[0]
                ancestor_count = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for _ in range(ancestor_count):
                    ancestor_pos = <uint16_t>hton.unpack_int16(
                        frb_read(spec, 2))
                    ancestor_codec = codecs_list[ancestor_pos]
            else:
                type_name = None
            pos = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            els = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            if els != 1:
                raise NotImplementedError(
                    'cannot handle arrays with more than one dimension')
            # First dimension length.
            dim_len = hton.unpack_int32(frb_read(spec, 4))
            sub_codec = <BaseCodec>codecs_list[pos]
            res = ArrayCodec.new(tid, sub_codec, dim_len)
            res.type_name = type_name

        elif t == CTYPE_RANGE:
            if protocol_version >= (2, 0):
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                type_name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                schema_defined = <bint>frb_read(spec, 1)[0]
                ancestor_count = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for _ in range(ancestor_count):
                    ancestor_pos = <uint16_t>hton.unpack_int16(
                        frb_read(spec, 2))
                    ancestor_codec = codecs_list[ancestor_pos]
            else:
                type_name = None
            pos = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            sub_codec = <BaseCodec>codecs_list[pos]
            res = RangeCodec.new(tid, sub_codec)
            res.type_name = type_name

        elif t == CTYPE_MULTIRANGE:
            if protocol_version >= (2, 0):
                str_len = hton.unpack_uint32(frb_read(spec, 4))
                type_name = cpythonx.PyUnicode_FromStringAndSize(
                    frb_read(spec, str_len), str_len)
                schema_defined = <bint>frb_read(spec, 1)[0]
                ancestor_count = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
                for _ in range(ancestor_count):
                    ancestor_pos = <uint16_t>hton.unpack_int16(
                        frb_read(spec, 2))
                    ancestor_codec = codecs_list[ancestor_pos]
            else:
                type_name = None
            pos = <uint16_t>hton.unpack_int16(frb_read(spec, 2))
            sub_codec = <BaseCodec>codecs_list[pos]
            res = MultiRangeCodec.new(tid, sub_codec)
            res.type_name = type_name

        elif t == CTYPE_OBJECT and protocol_version >= (2, 0):
            # Ignore
            frb_read(spec, desc_len)
            res = NULL_CODEC

        elif t == CTYPE_COMPOUND and protocol_version >= (2, 0):
            # Ignore
            frb_read(spec, desc_len)
            res = NULL_CODEC

        else:
            raise NotImplementedError(
                f'no codec implementation for EdgeDB data class {t}')

        self.codecs_build_cache[tid] = res
        return res

    cdef has_codec(self, bytes type_id):
        return (
            type_id in self.codecs or
            type_id in {NULL_CODEC_ID, EMPTY_TUPLE_CODEC_ID}
        )

    cdef BaseCodec get_codec(self, bytes type_id):
        try:
            return <BaseCodec>self.codecs[type_id]
        except KeyError:
            pass

        if type_id == NULL_CODEC_ID:
            return NULL_CODEC

        if type_id == EMPTY_TUPLE_CODEC_ID:
            return EMPTY_TUPLE_CODEC

        raise LookupError

    cdef BaseCodec build_codec(self, bytes spec, protocol_version):
        cdef:
            FRBuffer buf
            FRBuffer elem_buf
            BaseCodec res
            list codecs_list

        frb_init(
            &buf,
            cpython.PyBytes_AsString(spec),
            cpython.Py_SIZE(spec))

        codecs_list = []
        while frb_get_len(&buf):
            if protocol_version >= (2, 0):
                desc_len = <uint32_t>hton.unpack_int32(frb_read(&buf, 4))
                frb_slice_from(&elem_buf, &buf, desc_len)
                res = self._build_codec(
                    &elem_buf, codecs_list, protocol_version)
                if frb_get_len(&elem_buf):
                    raise RuntimeError(
                        f'unexpected trailing data in type descriptor datum')
            else:
                res = self._build_codec(&buf, codecs_list, protocol_version)
            if res is None:
                # An annotation; ignore.
                continue
            codecs_list.append(res)
            self.codecs[res.tid] = res

        if not codecs_list:
            raise RuntimeError(f'cannot not build codec; empty type desc')

        return codecs_list[-1]


cdef dict BASE_SCALAR_CODECS = {}


cdef register_base_scalar_codec(
        str name,
        pgproto.encode_func encoder,
        pgproto.decode_func decoder,
        object tid = None):

    cdef:
        BaseCodec codec

    if tid is None:
        tid = TYPE_IDS.get(name)
        if tid is None:
            raise RuntimeError(f'cannot find known ID for type {name!r}')
    tid = tid.bytes

    if tid in BASE_SCALAR_CODECS:
        raise RuntimeError(f'base scalar codec for {id} is already registered')

    codec = ScalarCodec.new(tid, name, encoder, decoder)
    BASE_SCALAR_CODECS[tid] = codec


cdef time_encode(pgproto.CodecContext settings, WriteBuffer buf, obj):
    if getattr(obj, 'tzinfo', None) is not None:
        raise TypeError(
            f'a naive time object (tzinfo is None) was expected')

    pgproto.time_encode(settings, buf, obj)


cdef date_encode(pgproto.CodecContext settings, WriteBuffer buf, obj):
    # Python `datetime.date` object does not have a `tzinfo` attribute
    # nor it is timezone-aware.  But since we're accepting duck types
    # let's ensure it doesn't have tzinfo anyways.
    if getattr(obj, 'tzinfo', None) is not None:
        raise TypeError(
            f'a naive date object (tzinfo is None) was expected')

    pgproto.date_encode(settings, buf, obj)


cdef timestamp_encode(pgproto.CodecContext settings, WriteBuffer buf, obj):
    if not cpython.datetime.PyDateTime_Check(obj):
        raise TypeError(
            f'a datetime.datetime object was expected, got {obj!r}')

    if getattr(obj, 'tzinfo', None) is not None:
        raise TypeError(
            f'a naive datetime object (tzinfo is None) was expected')

    pgproto.timestamp_encode(settings, buf, obj)


cdef timestamptz_encode(pgproto.CodecContext settings, WriteBuffer buf, obj):
    if not cpython.datetime.PyDateTime_Check(obj):
        raise TypeError(
            f'a datetime.datetime object was expected')

    if getattr(obj, 'tzinfo', None) is None:
        raise TypeError(
            f'a timezone-aware datetime object (tzinfo is not None) '
            f'was expected')

    pgproto.timestamptz_encode(settings, buf, obj)


cdef duration_encode(pgproto.CodecContext settings, WriteBuffer buf,
                     object obj):

    cdef microseconds = obj / datetime.timedelta(microseconds=1)

    buf.write_int32(16)
    buf.write_int64(microseconds)
    buf.write_int32(0)
    buf.write_int32(0)


cdef duration_decode(pgproto.CodecContext settings, FRBuffer *buf):
    cdef:
        int32_t days
        int32_t months
        int64_t microseconds

    microseconds = hton.unpack_int64(frb_read(buf, 8))
    days = hton.unpack_int32(frb_read(buf, 4))
    months = hton.unpack_int32(frb_read(buf, 4))
    if days != 0 or months != 0:
        raise RuntimeError(f'non-zero reserved bytes in duration')

    return datetime.timedelta(microseconds=microseconds)


cdef relative_duration_encode(pgproto.CodecContext settings, WriteBuffer buf,
                              object obj):

    cdef:
        microseconds = obj.microseconds
        days = obj.days
        months = obj.months

    buf.write_int32(16)
    buf.write_int64(microseconds)
    buf.write_int32(days)
    buf.write_int32(months)


cdef relative_duration_decode(pgproto.CodecContext settings, FRBuffer *buf):
    cdef:
        int32_t days
        int32_t months
        int64_t microseconds

    microseconds = hton.unpack_int64(frb_read(buf, 8))
    days = hton.unpack_int32(frb_read(buf, 4))
    months = hton.unpack_int32(frb_read(buf, 4))

    return datatypes.RelativeDuration(
        microseconds=microseconds, days=days, months=months)


cdef date_duration_encode(pgproto.CodecContext settings, WriteBuffer buf,
                          object obj):

    cdef:
        days = obj.days
        months = obj.months

    buf.write_int32(16)
    buf.write_int64(0)
    buf.write_int32(days)
    buf.write_int32(months)


cdef date_duration_decode(pgproto.CodecContext settings, FRBuffer *buf):
    cdef:
        int32_t days
        int32_t months
        int64_t microseconds

    microseconds = hton.unpack_int64(frb_read(buf, 8))
    days = hton.unpack_int32(frb_read(buf, 4))
    months = hton.unpack_int32(frb_read(buf, 4))

    if microseconds != 0:
        raise ValueError("date duration has non-zero microseconds")

    return datatypes.DateDuration(days=days, months=months)


cdef config_memory_encode(pgproto.CodecContext settings,
                          WriteBuffer buf,
                          object obj):
    cdef:
        bytes = obj._bytes

    buf.write_int32(8)
    buf.write_int64(bytes)


cdef config_memory_decode(pgproto.CodecContext settings, FRBuffer *buf):
    cdef:
        int64_t bytes

    bytes = hton.unpack_int64(frb_read(buf, 8))

    return datatypes.ConfigMemory(bytes=bytes)


DEF PGVECTOR_MAX_DIM = (1 << 16) - 1


cdef pgvector_encode_memview(pgproto.CodecContext settings, WriteBuffer buf,
                             float[:] obj):
    cdef:
        Py_ssize_t objlen
        Py_ssize_t i

    objlen = len(obj)
    if objlen > PGVECTOR_MAX_DIM:
        raise ValueError('too many elements in vector value')

    buf.write_int32(4 + objlen*4)
    buf.write_int16(objlen)
    buf.write_int16(0)
    for i in range(objlen):
        buf.write_float(obj[i])


cdef pgvector_encode(pgproto.CodecContext settings, WriteBuffer buf,
                     object obj):
    cdef:
        Py_ssize_t objlen
        float[:] memview
        Py_ssize_t i

    # If we can take a typed memview of the object, we use that.
    # That is good, because it means we can consume array.array and
    # numpy.ndarray without needing to unbox.
    # Otherwise we take the slow path, indexing into the array using
    # the normal protocol.
    try:
        memview = obj
    except (ValueError, TypeError) as e:
        pass
    else:
        pgvector_encode_memview(settings, buf, memview)
        return

    if not _is_array_iterable(obj):
        raise TypeError(
            'a sized iterable container expected (got type {!r})'.format(
                type(obj).__name__))

    # Annoyingly, this is literally identical code to the fast path...
    # but the types are different in critical ways.
    objlen = len(obj)
    if objlen > PGVECTOR_MAX_DIM:
        raise ValueError('too many elements in vector value')

    buf.write_int32(4 + objlen*4)
    buf.write_int16(objlen)
    buf.write_int16(0)
    for i in range(objlen):
        buf.write_float(obj[i])


cdef object ONE_EL_F32_ARRAY = array.array('f', [0.0])


cdef pgvector_decode(pgproto.CodecContext settings, FRBuffer *buf):
    cdef:
        int32_t dim
        Py_ssize_t size
        Py_buffer view
        char *p
        float[:] array_view

    dim = hton.unpack_uint16(frb_read(buf, 2))
    frb_read(buf, 2)

    size = dim * 4
    p = frb_read(buf, size)

    # Create a float array with size dim
    val = ONE_EL_F32_ARRAY * dim

    # And fill it with the buffer contents
    array_view = val
    memcpy(&array_view[0], p, size)
    val.byteswap()

    return val


# Half-floats or float16 types often don't have a native implementation, so in
# order to cast to and from it we need to implement the cast ourselves. The
# algorithm to do so is taken from here:
# http://www.fox-toolkit.org/ftp/fasthalffloatconversion.pdf
#
# It is a paper titled "Fast Half Float Conversions" by Jeroen van der Zijp,
# referenced by wiki as well as many articles on float conversion.

# Needed to perform bit manipulations to convert from float16 to float32
cdef union IntFloatSwap:
    uint32_t i
    float f


# Optimization to convert f16 to f32
cdef:
    uint32_t[2048] MANTISSA_TABLE
    uint32_t[64] EXP_TABLE
    uint32_t[64] OFFSET_TABLE
    uint32_t[512] BASE_TABLE
    uint32_t[512] SHIFT_TABLE


cdef convertmantissa(uint32_t n):
    cdef:
        uint32_t m
        uint32_t e

    m = n << 13                 # Zero pad mantissa bits
    e = 0                       # Zero exponent
    # While not normalized
    while (m & 0x00800000) == 0:
        e -= 0x00800000         # Decrement exponent (1<<23)
        m <<= 1                 # Shift mantissa

    m &= ~0x00800000            # Clear leading 1 bit
    e += 0x38800000             # Adjust bias ((127-14)<<23)

    return m | e


cdef generate_tables():
    cdef:
        uint32_t i
        int32_t e

    # Generate the f16 to f32 conversion tables
    MANTISSA_TABLE[0] = 0
    for i in range(1, 1024):
        MANTISSA_TABLE[i] = convertmantissa(i)
    for i in range(1024, 2048):
        MANTISSA_TABLE[i] = 0x38000000 + ((i - 1024) << 13)

    EXP_TABLE[0] = 0
    EXP_TABLE[31]= 0x47800000
    EXP_TABLE[32]= 0x80000000
    EXP_TABLE[63]= 0xC7800000
    for i in range(1, 31):
        EXP_TABLE[i] = i << 23
    for i in range(33, 63):
        EXP_TABLE[i] = 0x80000000 + ((i - 32) << 23)

    for i in range(64):
        if i == 0 or i == 32:
            OFFSET_TABLE[i] = 0
        else:
            OFFSET_TABLE[i] = 1024

    # Generate the f32 to f16 conversion tables
    for i in range(256):
        e = i - 127

        # Very small numbers map to zero
        if e < -24:
            BASE_TABLE[i] = 0
            BASE_TABLE[i + 256] = 0x8000
            SHIFT_TABLE[i] = 24
            SHIFT_TABLE[i + 256] = 24

        # Small numbers map to denorms
        elif e < -14:
            BASE_TABLE[i] = 0x0400 >> (-e - 14)
            BASE_TABLE[i + 256] = (0x0400 >> (-e - 14)) | 0x8000
            SHIFT_TABLE[i] = -e - 1
            SHIFT_TABLE[i + 256] = -e - 1

        # Normal numbers just lose precision
        elif e <= 15:
            BASE_TABLE[i] = (e + 15) << 10
            BASE_TABLE[i + 256] = ((e + 15) << 10) | 0x8000
            SHIFT_TABLE[i] = 13
            SHIFT_TABLE[i + 256] = 13

        # Large numbers map to Infinity
        elif e < 128:
            BASE_TABLE[i] = 0x7C00
            BASE_TABLE[i + 256] = 0xFC00
            SHIFT_TABLE[i] = 24
            SHIFT_TABLE[i + 256] = 24

        # Infinity and NaN's stay Infinity and NaN's
        else:
            BASE_TABLE[i] = 0x7C00
            BASE_TABLE[i + 256] = 0xFC00
            SHIFT_TABLE[i] = 13
            SHIFT_TABLE[i + 256] = 13


generate_tables()


cdef pgvector_hv_encode(pgproto.CodecContext settings, WriteBuffer buf,
                        object obj):
    cdef:
        Py_ssize_t objlen
        float[:] memview
        Py_ssize_t i
        IntFloatSwap swap

    if not _is_array_iterable(obj):
        raise TypeError(
            'a sized iterable container expected (got type {!r})'.format(
                type(obj).__name__))

    objlen = len(obj)
    if objlen > PGVECTOR_MAX_DIM:
        raise ValueError('too many elements in vector value')

    buf.write_int32(4 + objlen * 2)
    buf.write_int16(objlen)
    buf.write_int16(0)
    for i in range(objlen):
        swap.f = obj[i]
        swap.i = (
            BASE_TABLE[(swap.i >> 23) & 0x1ff]
            + ((swap.i & 0x007fffff) >> SHIFT_TABLE[(swap.i >> 23) & 0x1ff])
        )

        if swap.i == BASE_TABLE[255] or swap.i == BASE_TABLE[511]:
            raise ValueError(
                '{!r} is out of range for type halfvec'.format(obj[i]))

        buf.write_int16(swap.i)


cdef object ONE_EL16_ARRAY = array.array('H', [0])


cdef pgvector_hv_decode(pgproto.CodecContext settings, FRBuffer *buf):
    cdef:
        int32_t dim
        Py_ssize_t size
        Py_buffer view
        char *p
        unsigned short[:] tmp_array_view
        float[:] array_view
        Py_ssize_t i
        IntFloatSwap swap

    dim = hton.unpack_uint16(frb_read(buf, 2))
    frb_read(buf, 2)

    # We'll read float16 values as an array of uint16 and then convert since
    # plain vanilla Python doesn't have half-floats.
    #
    # TODO: Potentially we might handle it differently if numpy is present since
    # numpy has half floats since 1.60.
    size = dim * 2
    p = frb_read(buf, size)
    tmp = ONE_EL16_ARRAY * dim

    # And fill it with the buffer contents
    tmp_array_view = tmp
    memcpy(&tmp_array_view[0], p, size)
    tmp.byteswap()

    # Create a float array with size dim
    val = ONE_EL_F32_ARRAY * dim
    for i in range(dim):
        swap.i = tmp[i]
        swap.i = MANTISSA_TABLE[
            OFFSET_TABLE[swap.i >> 10] + (swap.i & 0x000003ff)
        ] + EXP_TABLE[swap.i >> 10]
        val[i] = swap.f

    return val


cdef pgvector_sv_encode(pgproto.CodecContext settings, WriteBuffer buf,
                        object obj):
    cdef:
        int32_t dim
        Py_ssize_t nnz

    if not (cpython.PyDict_Check(obj) or isinstance(obj, MappingABC)):
        raise TypeError(
            'a dict or Mapping expected (got type {!r})'.format(
                type(obj).__name__))

    if 'dim' not in obj:
        raise TypeError('"dim" key is missing')

    dim = obj['dim']

    # One of the keys is 'dim' so nnz is one less than the length
    nnz = len(obj) - 1
    if nnz > PGVECTOR_MAX_DIM or nnz > dim:
        raise ValueError('too many elements in vector value')

    buf.write_int32(4 * (3 + 2 * nnz))
    buf.write_int32(dim)
    buf.write_int32(nnz)
    buf.write_int32(0)

    for key in obj.keys():
        if key != 'dim':
            ensure_is_int(key)
            buf.write_int32(key)

    for key, val in obj.items():
        if key != 'dim':
            if val == 0:
                raise ValueError('elements in a sparse vector cannot be 0')
            buf.write_float(<float>val)


cdef object ONE_EL_I32_ARRAY = array.array('l', [0])


cdef pgvector_sv_decode(pgproto.CodecContext settings, FRBuffer *buf):
    cdef:
        int32_t dim
        int32_t nnz
        Py_ssize_t size
        char *p
        float[:] array_view
        Py_ssize_t i

    dim = hton.unpack_int32(frb_read(buf, 4))
    nnz = hton.unpack_int32(frb_read(buf, 4))
    frb_read(buf, 4)

    # Create an int array with size nnz (for indexes)
    keys = ONE_EL_I32_ARRAY * nnz
    for i in range(nnz):
        keys[i] = hton.unpack_int32(frb_read(buf, 4))

    size = nnz * 4
    p = frb_read(buf, size)

    # Create a float array with size nnz (for values)
    vals = ONE_EL_F32_ARRAY * nnz

    # And fill it with the buffer contents
    array_view = vals
    memcpy(&array_view[0], p, size)
    vals.byteswap()

    res = {'dim': dim}
    res.update(zip(keys, vals))

    return res


cdef checked_decimal_encode(
    pgproto.CodecContext settings, WriteBuffer buf, obj
):
    if not isinstance(obj, decimal.Decimal) and not isinstance(obj, int):
        raise TypeError('expected a Decimal or an int')
    pgproto.numeric_encode_binary(settings, buf, obj)


cdef decimal_decode(pgproto.CodecContext settings, FRBuffer *buf):
    return pgproto.numeric_decode_binary_ex(settings, buf, True)


cdef ensure_is_int(obj):
    if type(obj) is not int and not isinstance(obj, int):
        raise TypeError('expected an int')

cdef checked_int2_encode(pgproto.CodecContext settings, WriteBuffer buf, obj):
    ensure_is_int(obj)
    pgproto.int2_encode(settings, buf, obj)


cdef checked_int4_encode(pgproto.CodecContext settings, WriteBuffer buf, obj):
    ensure_is_int(obj)
    pgproto.int4_encode(settings, buf, obj)


cdef checked_int8_encode(pgproto.CodecContext settings, WriteBuffer buf, obj):
    ensure_is_int(obj)
    pgproto.int8_encode(settings, buf, obj)


cdef bigint_encode(pgproto.CodecContext settings, WriteBuffer buf, obj):
    cdef:
        uint16_t sign = NUMERIC_POS
        list digits = []

    ensure_is_int(obj)

    if obj == 0:
        buf.write_int32(8)  # len
        buf.write_int32(0)  # ndigits + weight
        buf.write_int16(NUMERIC_POS)  # sign
        buf.write_int16(0)  # dscale
        return

    if obj < 0:
        sign = NUMERIC_NEG
        obj = -obj

    while obj:
        mod = obj % NBASE
        obj //= NBASE
        digits.append(mod)

    buf.write_int32(8 + len(digits) * 2)  # len
    buf.write_int16(len(digits))  # ndigits
    buf.write_int16(len(digits) - 1)  # weight
    buf.write_int16(sign)  # sign
    buf.write_int16(0)  # dscale
    for dig in reversed(digits):
        buf.write_int16(dig)


cdef bigint_decode(pgproto.CodecContext settings, FRBuffer *buf):
    cdef:
        uint16_t ndigits = <uint16_t>hton.unpack_int16(frb_read(buf, 2))
        uint16_t weight = <uint16_t>hton.unpack_int16(frb_read(buf, 2))
        uint16_t sign = <uint16_t>hton.unpack_int16(frb_read(buf, 2))
        uint16_t dscale = <uint16_t>hton.unpack_int16(frb_read(buf, 2))
        result = ''
        int32_t i = <int32_t>weight
        uint16_t d = 0

    if sign == NUMERIC_NEG:
        result = '-'
    elif sign != NUMERIC_POS:
        raise ValueError("bad bigint sign data")

    if dscale != 0:
        raise ValueError("bigint data has fractional part")

    if ndigits == 0:
        return 0

    while i >= 0:
        if i <= weight and d < ndigits:
            num = str(<uint16_t>hton.unpack_int16(frb_read(buf, 2)))
            result += num.zfill(4)
            d += 1
        else:
            result += '0000'
        i -= 1

    return int(result)


cdef geometry_encode(pgproto.CodecContext settings, WriteBuffer buf, obj):
    buf.write_int32(len(obj.wkb))
    buf.write_bytes(obj.wkb)


cdef geometry_decode(pgproto.CodecContext settings, FRBuffer *buf):
    cdef:
        object result
        ssize_t buf_len

    # Just wrap the bytes into a named tuple with a single field `wkb`
    descriptor = datatypes.record_desc_new(
        ('wkb',), <object>NULL, <object>NULL)
    result = datatypes.namedtuple_new(
        datatypes.namedtuple_type_new(descriptor))

    # frb_read_all adjusts buf.len, so we need to read it out first
    buf_len = buf.len
    elem = PyBytes_FromStringAndSize(frb_read_all(buf), buf_len)
    cpython.Py_INCREF(elem)
    cpython.PyTuple_SET_ITEM(result, 0, elem)

    return result


cdef register_base_scalar_codecs():
    register_base_scalar_codec(
        'std::uuid',
        pgproto.uuid_encode,
        pgproto.uuid_decode)

    register_base_scalar_codec(
        'std::str',
        pgproto.text_encode,
        pgproto.text_decode)

    register_base_scalar_codec(
        'std::bytes',
        pgproto.bytea_encode,
        pgproto.bytea_decode)

    register_base_scalar_codec(
        'std::int16',
        checked_int2_encode,
        pgproto.int2_decode)

    register_base_scalar_codec(
        'std::int32',
        checked_int4_encode,
        pgproto.int4_decode)

    register_base_scalar_codec(
        'std::int64',
        checked_int8_encode,
        pgproto.int8_decode)

    register_base_scalar_codec(
        'std::float32',
        pgproto.float4_encode,
        pgproto.float4_decode)

    register_base_scalar_codec(
        'std::float64',
        pgproto.float8_encode,
        pgproto.float8_decode)

    register_base_scalar_codec(
        'std::decimal',
        checked_decimal_encode,
        decimal_decode)

    register_base_scalar_codec(
        'std::bigint',
        bigint_encode,
        bigint_decode)

    register_base_scalar_codec(
        'std::bool',
        pgproto.bool_encode,
        pgproto.bool_decode)

    register_base_scalar_codec(
        'std::datetime',
        timestamptz_encode,
        pgproto.timestamptz_decode)

    register_base_scalar_codec(
        'std::pg::timestamptz',
        timestamptz_encode,
        pgproto.timestamptz_decode)

    register_base_scalar_codec(
        'cal::local_datetime',
        timestamp_encode,
        pgproto.timestamp_decode)

    register_base_scalar_codec(
        'std::pg::timestamp',
        timestamp_encode,
        pgproto.timestamp_decode)

    register_base_scalar_codec(
        'cal::local_date',
        date_encode,
        pgproto.date_decode)

    register_base_scalar_codec(
        'std::pg::date',
        date_encode,
        pgproto.date_decode)

    register_base_scalar_codec(
        'cal::local_time',
        time_encode,
        pgproto.time_decode)

    register_base_scalar_codec(
        'std::duration',
        duration_encode,
        duration_decode)

    register_base_scalar_codec(
        'std::json',
        pgproto.jsonb_encode,
        pgproto.jsonb_decode)

    register_base_scalar_codec(
        'std::pg::json',
        pgproto.json_encode,
        pgproto.json_decode)

    register_base_scalar_codec(
        'cal::relative_duration',
        relative_duration_encode,
        relative_duration_decode)

    register_base_scalar_codec(
        'std::pg::interval',
        relative_duration_encode,
        relative_duration_decode)

    register_base_scalar_codec(
        'cal::date_duration',
        date_duration_encode,
        date_duration_decode)

    register_base_scalar_codec(
        'cfg::memory',
        config_memory_encode,
        config_memory_decode)

    register_base_scalar_codec(
        'fts::language',
        pgproto.text_encode,
        pgproto.text_decode)

    register_base_scalar_codec(
        'ext::pgvector::vector',
        pgvector_encode,
        pgvector_decode,
        uuid.UUID('9565dd88-04f5-11ee-a691-0b6ebe179825'),
    )

    register_base_scalar_codec(
        'ext::pgvector::halfvec',
        pgvector_hv_encode,
        pgvector_hv_decode,
        uuid.UUID('4ba84534-188e-43b4-a7ce-cea2af0f405b'),
    )

    register_base_scalar_codec(
        'ext::pgvector::sparsevec',
        pgvector_sv_encode,
        pgvector_sv_decode,
        uuid.UUID('003e434d-cac2-430a-b238-fb39d73447d2'),
    )

    register_base_scalar_codec(
        'ext::postgis::geometry',
        geometry_encode,
        geometry_decode,
        uuid.UUID('44c901c0-d922-4894-83c8-061bd05e4840'),
    )

    register_base_scalar_codec(
        'ext::postgis::geography',
        geometry_encode,
        geometry_decode,
        uuid.UUID('4d738878-3a5f-4821-ab76-9d8e7d6b32c4'),
    )

    register_base_scalar_codec(
        'ext::postgis::box2d',
        geometry_encode,
        geometry_decode,
        uuid.UUID('7fae5536-6311-4f60-8eb9-096a5d972f48'),
    )

    register_base_scalar_codec(
        'ext::postgis::box3d',
        geometry_encode,
        geometry_decode,
        uuid.UUID('c1a50ff8-fded-48b0-85c2-4905a8481433'),
    )


register_base_scalar_codecs()



#
# This source file is part of the EdgeDB open source project.
#
# Copyright 2022-present MagicStack Inc. and the EdgeDB authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import argparse
import sys

from . import generator


class ColoredArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        c = generator.C
        self.exit(
            2,
            f"{c.BOLD}{c.FAIL}error:{c.ENDC} "
            f"{c.BOLD}{message:s}{c.ENDC}\n",
        )


parser = ColoredArgumentParser(
    description="Generate Python code for .edgeql files."
)
parser.add_argument("--dsn")
parser.add_argument("--credentials-file", metavar="PATH")
parser.add_argument("-I", "--instance", metavar="NAME")
parser.add_argument("-H", "--host")
parser.add_argument("-P", "--port")
parser.add_argument("-d", "--database", metavar="NAME")
parser.add_argument("-u", "--user")
parser.add_argument("--password")
parser.add_argument("--password-from-stdin", action="store_true")
parser.add_argument("--tls-ca-file", metavar="PATH")
parser.add_argument(
    "--tls-security",
    choices=["default", "strict", "no_host_verification", "insecure"],
)
parser.add_argument(
    "--file",
    action="append",
    nargs="?",
    help="Generate a single file instead of one per .edgeql file.",
)
parser.add_argument(
    "--dir",
    action="append",
    help="Only search .edgeql files under specified directories.",
)
parser.add_argument(
    "--target",
    choices=["blocking", "async"],
    nargs="*",
    default=["async"],
    help="Choose one or more targets to generate code (default is async)."
)
if sys.version_info[:2] >= (3, 9):
    parser.add_argument(
        "--skip-pydantic-validation",
        action=argparse.BooleanOptionalAction,
        default=argparse.SUPPRESS,  # override the builtin help for default
        help="Add a mixin to generated dataclasses "
        "to skip Pydantic validation (default is to add the mixin).",
    )
else:
    parser.add_argument(
        "--skip-pydantic-validation",
        action="store_true",
        default=True,
    )
    parser.add_argument(
        "--no-skip-pydantic-validation",
        dest="skip_pydantic_validation",
        action="store_false",
        default=False,
        help="Add a mixin to generated dataclasses "
             "to skip Pydantic validation (default is to add the mixin).",
    )


def main():
    args = parser.parse_args()
    if not hasattr(args, "skip_pydantic_validation"):
        args.skip_pydantic_validation = True
    generator.Generator(args).run()



import unittest

# No tests here, but we want to skip the unittest loader from attempting to
# import ORM packages which may not have been installed (like Django that has
# a few custom adjustments to make our models work).
def load_tests(loader, tests, pattern):
    return tests
